'''module to interrogate equations to withdraw essential information for solution'''
# code to extract the equation information for chemical reactions required for 
# their solution in PyCHAM 

import numpy as np
import re
import formatting
import pybel
import sys

def eqn_interr(num_eqn, eqn_list, aqeqn_list, chem_scheme_markers, spec_name, 
		spec_smil, num_sb, wall_on):
				
	# inputs: ----------------------------------------------------------------------------
	# num_eqn - number of equations
	# eqn_list - gas-phase equations in strings
	# aqeqn_list - aqueous-phase equations in strings
	# chem_scheme_markers - markers for separating sections of the chemical scheme
	# spec_name - name string of components in xml file (not SMILES)
	# spec_smil - SMILES from xml file
	# num_sb - number of size bins
	# wall_on - marker for whether to include wall partitioning
	# ------------------------------------------------------------------------------------
	
	# preparatory part ----------------------------------------------------
	# matrix to record indices of reactants (cols) in each equation (rows)
	rindx = np.zeros((num_eqn[0], 1)).astype(int)
	# matrix to arrange concentrations when reaction rate coefficient calculated
	y_arr = (np.ones((num_eqn[0], 1)).astype(int))*-9999
	# array to arrange reaction rates so they align with reactant stoichiometries
	rr_arr = np.empty((0))
	# same but for products
	rr_arr_p = np.empty((0))
	# index array for extracting required reactant concentrations for the
	# reaction rate coefficient calculation
	y_rind = np.empty((0))
	# index array for identifying products when assigning gains from reactions
	y_pind = np.empty((0))	
	# matrix to record indices of products (cols) in each equation (rows)
	pindx = np.zeros((num_eqn[0], 1)).astype(int)
	# matrix to record stoichiometries of reactants (cols) in each equation (rows)
	rstoi = np.ones((num_eqn[0], 1))
	jac_stoi = np.zeros((num_eqn[0], 1))
	# 1D array to record stoichiometries of reactants per equarion
	rstoi_flat = np.empty((0))
	# 1D array to record stoichiometries of products per equarion
	pstoi_flat = np.empty((0))
	# matrix to record stoichiometries of products (cols) in each equation (rows)
	pstoi = np.zeros((num_eqn[0], 1))
	# arrays to store number of reactants and products in gas-phase equations
	nreac = np.empty(num_eqn[0], dtype=np.int8)
	nprod = np.empty(num_eqn[0], dtype=np.int8)
	# list for equation reaction rate coefficients
	reac_coef = []
	# matrix containing index of components who are denominators in the
	# calculation of equation derivatives in the Jacobian
	jac_den_indx = np.zeros((num_eqn[0], 1))
	# total number of Jacobian elements per equation
	njac = np.zeros((num_eqn[0], 1))
	# indices of Jacobian to affect per equation (rows)
	jac_indx = np.zeros((num_eqn[0], 1))
	# a new list for the name strings of components presented in the scheme (not SMILES)
	comp_namelist = []
	# list for the SMILE strings of components present in the chemical scheme
	comp_list = []
	# list of Pybel objects of components in chemical scheme
	Pybel_objects = []
	comp_num = 0 # count the number of unique components in the chemical scheme
	# ---------------------------------------------------------------------

	max_no_reac = 0.0 # log maximum number of reactants in a reaction
	max_no_prod = 0.0 # log maximum number of products in a reaction

	# loop through gas-phase equations line by line and extract the required information
	for eqn_step in range(num_eqn[0]):
		
		line = eqn_list[eqn_step] # extract this line
		
		# work out whether equation or reaction rate coefficient part comes first
		eqn_start = str('.*\\' +  chem_scheme_markers[10])
		rrc_start = str('.*\\' +  chem_scheme_markers[9])
		# get index of these markers, note span is the property of the match object that
		# gives the location of the marker
		eqn_start_indx = (re.match(eqn_start, line)).span()[1]
		rrc_start_indx = (re.match(rrc_start, line)).span()[1]
		
		if eqn_start_indx>rrc_start_indx:
			eqn_sec = 1 # equation is second part
		else:
			eqn_sec = 0 # equation is first part
		
		# split the line into 2 parts: equation and rate coefficient
		# . means match with anything except a new line character., when followed by a * 
		# means match zero or more times (so now we match with all characters in the line
		# except for new line characters, so final part is stating the character(s) we 
		# are specifically looking for, \\ ensures the marker is recognised
		if eqn_sec == 1:
			eqn_markers = str('\\' +  chem_scheme_markers[10]+ '.*\\' +  chem_scheme_markers[11])
		else: # end of equation part is start of reaction rate coefficient part
			eqn_markers = str('\\' +  chem_scheme_markers[10]+ '.*\\' +  chem_scheme_markers[9])

		# extract the equation as a string ([0] extracts the equation section and 
		# [1:-1] removes the bounding markers)
		eqn = re.findall(eqn_markers, line)[0][1:-1].strip()
		
		eqn_split = eqn.split()
		eqmark_pos = eqn_split.index('=')
		# with stoich number; rule out the photon
		reactants = [i for i in eqn_split[:eqmark_pos] if i != '+' and i != 'hv']
		products = [t for t in eqn_split[eqmark_pos+1:] if t != '+'] # with stoich number
		
		# record maximum number of reactants across all equations
		max_no_reac = np.maximum(len(reactants), max_no_reac)
		# record maximum number of products across all equations
		max_no_prod = np.maximum(len(products), max_no_prod)

		# append columns if needed
		while max_no_reac > np.minimum(rindx.shape[1], rstoi.shape[1]): 
			rindx = np.append(rindx, (np.zeros((num_eqn[0], 1))).astype(int), axis=1)
			rstoi = np.append(rstoi, (np.ones((num_eqn[0], 1))), axis=1)
			y_arr = np.append(y_arr, (np.ones((num_eqn[0], 1))*-9999).astype(int), axis=1)
			y_arr_fixer = ((np.arange(0, num_eqn[0], dtype = 'int')).reshape(-1, 1)).repeat(max_no_reac, axis=1)
			y_arr[y_arr!=-9999] = y_arr[y_arr!=-9999]+y_arr_fixer[y_arr!=-9999] 
		while max_no_prod > np.minimum(pindx.shape[1], pstoi.shape[1]): 
			pindx = np.append(pindx, (np.zeros((num_eqn[0], 1))).astype(int), axis=1)
			pstoi = np.append(pstoi, (np.zeros((num_eqn[0], 1))), axis=1)
		while ((len(reactants)**2.0+len(reactants)*len(products))>jac_indx.shape[1]):
			jac_indx = np.append(jac_indx, (np.zeros((num_eqn[0], 1))), axis=1)
			jac_den_indx = np.append(jac_den_indx, (np.zeros((num_eqn[0], 1))), axis=1)		
			jac_stoi = np.append(jac_stoi, (np.zeros((num_eqn[0], 1))), axis=1)

		# .* means occurs anywhere in line and, first \ means second \ can be interpreted 
		# and second \ ensures recognition of marker
		rate_coeff_start_mark = str('\\' +  chem_scheme_markers[9])
		# . means match with anything except a new line character, when followed by a * 
		# means match zero or more times (so now we match with all characters in the line
		# except for new line characters, \\ ensures the marker
		# is recognised
		if eqn_sec == 1: # end of reaction rate coefficient part is start of equation part
			rate_coeff_end_mark = str('.*\\' +  chem_scheme_markers[10])
		else: # end of reaction rate coefficient part is end of line
			rate_coeff_end_mark = str('.*\\' +  chem_scheme_markers[11])
		
		# rate coefficient starts and end punctuation
		rate_regex = str(rate_coeff_start_mark + rate_coeff_end_mark)
		# rate coefficient expression in a string
		rate_ex = re.findall(rate_regex, line)[0][1:-1].strip()

		# convert fortran-type scientific notation to python type
		rate_ex = formatting.SN_conversion(rate_ex)
		# convert the rate coefficient expressions into Python readable commands
		rate_ex = formatting.convert_rate_mcm(rate_ex)
		if (rate_ex.find('EXP') != -1):
			print('Error in reaction rate coefficient expression: ', rate_ex)
			sys.exit()
		
		# store the reaction rate coefficient for this equation 
		# (/s once any inputs applied)
		reac_coef.append(rate_ex)
		
		# extract the stoichiometric number of the component in current equation
		reactant_step = 0
		product_step = 0
		stoich_regex = r"^\d*\.\d*|^\d*"
		numr = len(reactants) # number of reactants in this equation
		
		
		# left hand side of equations (losses)
		for reactant in reactants:
				
			if (re.findall(stoich_regex, reactant)[0] != ''):
				stoich_num = float(re.findall(stoich_regex, reactant)[0])
				# name with no stoich number
				name_only = re.sub(stoich_regex, '', reactant)
			elif (re.findall(stoich_regex, reactant)[0] == ''):
				stoich_num = 1.0
				name_only = reactant
			
			# store stoichiometry
			rstoi[eqn_step, reactant_step] = stoich_num
			jac_stoi[eqn_step, reactant_step] = -1*stoich_num

			if name_only not in comp_namelist: # if new component encountered
				comp_namelist.append(name_only) # add to chemical scheme name list
			
				# convert MCM chemical names to SMILES
				if (name_only in spec_name):
					# index where xml file name matches reaction component name
					name_indx = spec_name.index(name_only)
					name_SMILE = spec_smil[name_indx] # SMILES of component
				else:
					print(str('Error: inside eqn_parser, chemical scheme name '+str(name_only)+' not found in xml file'))
					sys.exit()
			
				comp_list.append(name_SMILE) # list SMILE names
				name_indx = comp_num # allocate index to this species
				# generate pybel object
				Pybel_object = pybel.readstring('smi', name_SMILE)
				# append to Pybel object list
				Pybel_objects.append(Pybel_object)
				
				comp_num += 1 # number of unique species
				

			else: # if it's a species already encountered it will be in comp_list
				# existing index
				name_indx = comp_namelist.index(name_only)
			
			# store reactant index
			# check if index already present - i.e. component appears more than once
			if sum(rindx[eqn_step, 0:reactant_step]==int(name_indx))>0:
				# get existing index of this component
				exist_indx = (np.where(rindx[eqn_step, 0:reactant_step]==(int(name_indx))))[0]
				# add to existing stoichiometry
				rstoi[eqn_step, exist_indx] += rstoi[eqn_step, reactant_step]
				jac_stoi[eqn_step, exist_indx] += -1*rstoi[eqn_step, reactant_step]
				# remove stoichiometry added above
				rstoi[eqn_step, reactant_step] = 1
				jac_stoi[eqn_step, reactant_step] = 0
				reactant_step -= 1 # ignore this duplicate
			else:
				rindx[eqn_step, reactant_step] = int(name_indx)
				y_arr[eqn_step, reactant_step] = int((eqn_step*max_no_reac)+reactant_step)
				y_rind = np.append(y_rind, int(name_indx))
				rr_arr = np.append(rr_arr, int(eqn_step))
		
			reactant_step += 1
					
		# number of reactants in this equation
		nreac[eqn_step] = int(reactant_step)
		
		# record 1D array of stoichiometries per equation
		rstoi_flat = np.append(rstoi_flat, rstoi[eqn_step, 0:int(reactant_step)])
		
		# right hand side of equations (gains)
		for product in products:

			if (re.findall(stoich_regex, product)[0] != ''):
				stoich_num = float(re.findall(stoich_regex, product)[0])
				name_only = re.sub(stoich_regex, '', product) # name with no stoich number

			elif (re.findall(stoich_regex, product)[0] == ''):
				stoich_num = 1.0
				name_only = product
			
			# store stoichiometry
			pstoi[eqn_step, product_step] = stoich_num
			jac_stoi[eqn_step, reactant_step+product_step] = 1*stoich_num
			if name_only not in comp_namelist: # if new component encountered
				comp_namelist.append(name_only)
				
				# convert MCM chemical names to SMILES
				# index where xml file name matches reaction component name
				if name_only in spec_name:
					name_indx = spec_name.index(name_only)
					name_SMILE = spec_smil[name_indx]
				else:
					print('Error: inside eqn_interr, chemical scheme name '+str(name_only)+' not found in xml file')
					sys.exit()
				
				comp_list.append(name_SMILE) # list SMILE string of parsed species
				name_indx = comp_num # allocate index to this species
				# Generate pybel
				
				Pybel_object = pybel.readstring('smi', name_SMILE)
				# append to Pybel object list
				Pybel_objects.append(Pybel_object)
				
				comp_num += 1 # number of unique species
				
			

			else: # if it's a species already encountered
				# index of component already listed
				name_indx = comp_namelist.index(name_only)
				
			# store product index
			# check if index already present - i.e. component appears more than once
			if sum(pindx[eqn_step, 0:product_step]==int(name_indx))>0:
				# get existing index of this component
				exist_indx = (np.where(pindx[eqn_step, 0:product_step]==(int(name_indx))))[0]
				# add to existing stoichiometry
				pstoi[eqn_step, exist_indx] += pstoi[eqn_step, product_step]
				jac_stoi[eqn_step, reactant_step+exist_indx] += 1*pstoi[eqn_step, product_step]
				# remove stoichiometry added above
				pstoi[eqn_step, product_step] = 0
				jac_stoi[eqn_step, reactant_step+product_step] = 0
				product_step -= 1 # ignore this duplicate
			else:
				pindx[eqn_step, product_step] = int(name_indx)
				rr_arr_p = np.append(rr_arr_p, int(eqn_step))
				y_pind = np.append(y_pind, int(name_indx))

			product_step += 1
		
		# number of products in this equation
		nprod[eqn_step] = int(product_step)
		# record 1D array of stoichiometries per equation
		pstoi_flat = np.append(pstoi_flat, pstoi[eqn_step, 0:int(product_step)])
		
		# now that total number of components (reactants and products) 
		# in an equation is known, repeat the reactant indices over all 
		# components
		tot_comp = nreac[eqn_step]+nprod[eqn_step]
		for i in range(nreac[eqn_step]):
			jac_den_indx[eqn_step, i*tot_comp:(i+1)*tot_comp] = rindx[eqn_step, i]
			# also repeat the stoichiometries for every reactant
			if (i > 0):
				jac_stoi[eqn_step, i*tot_comp:(i+1)*tot_comp] = jac_stoi[eqn_step, 0:tot_comp] 
 		# number of Jacobian elements affected by this equation
		njac[eqn_step, 0] = tot_comp*nreac[eqn_step]
	
	# flatten index for arranging concentrations ready for reaction rate coefficient calculation
	y_arr_g = y_arr.flatten(order='C')
	y_arr_g = y_arr[y_arr != -9999] # remove fillers
	y_rind_g = y_rind.astype(int) # ensure integer type
	uni_y_rind_g = (np.unique(y_rind)).astype(int) # unique index of reactants
	y_pind_g = y_pind.astype(int) # ensure integer type
	uni_y_pind_g = (np.unique(y_pind)).astype(int) # unique index of products
	rr_arr_g = rr_arr.astype(int) # ensure integer type
	rr_arr_p_g = rr_arr_p.astype(int) # ensure integer type	
	# colptrs for sparse matrix of the change to reactants per equation
	reac_col_g = np.cumsum(nreac)-nreac
	# include final column
	reac_col_g = np.append(reac_col_g, reac_col_g[-1]+nreac[-1])
	# colptrs for sparse matrix of the change to products per equation
	prod_col_g = np.cumsum(nprod)-nprod
	# include final column
	prod_col_g = np.append(prod_col_g, prod_col_g[-1]+nprod[-1])
	# tag other gas-phase arrays
	rindx_g = rindx
	pindx_g = pindx
	rstoi_g = rstoi
	pstoi_g = pstoi
	jac_stoi_g = jac_stoi
	rstoi_flat_g = rstoi_flat
	pstoi_flat_g = pstoi_flat
	nreac_g = nreac
	nprod_g = nprod
	reac_coef_g = reac_coef
	jac_den_indx_g = jac_den_indx.astype(int)
	njac_g = njac.astype(int)
	jac_indx_g = jac_indx
	jac_indx_g = jac_indx_g.astype(int)

	# repeat for aqueous-phase reactions ----------------------------------
	# preparatory part ----------------------------------------------------
	# matrix to record indices of reactants (cols) in each equation (rows)
	rindx = np.zeros((num_eqn[1], 1)).astype(int)
	# matrix to arrange concentrations when reaction rate coefficient calculated
	y_arr = (np.ones((num_eqn[1], 1)).astype(int))*-9999
	# array to arrange reaction rates so they align with reactant stoichiometries
	rr_arr = np.empty((0))
	# same but for products
	rr_arr_p = np.empty((0))
	# index array for extracting required reactant concentrations for the
	# reaction rate coefficient calculation
	y_rind = np.empty((0))
	# index array for identifying products when assigning gains from reactions
	y_pind = np.empty((0))	
	# matrix to record indices of products (cols) in each equation (rows)
	pindx = np.zeros((num_eqn[1], 1)).astype(int)
	# matrix to record stoichiometries of reactants (cols) in each equation (rows)
	rstoi = np.ones((num_eqn[1], 1))
	jac_stoi = np.zeros((num_eqn[1], 1))
	# 1D array to record stoichiometries of reactants per equarion
	rstoi_flat = np.empty((0))
	# 1D array to record stoichiometries of products per equarion
	pstoi_flat = np.empty((0))
	# matrix to record stoichiometries of products (cols) in each equation (rows)
	pstoi = np.zeros((num_eqn[1], 1))
	# arrays to store number of reactants and products of equations
	nreac = np.empty(num_eqn[1], dtype=np.int8)
	nprod = np.empty(num_eqn[1], dtype=np.int8)
	# list for equation reaction rate coefficients
	reac_coef = []
	# matrix containing index of components who are denominators in the
	# calculation of equation derivatives in the Jacobian
	jac_den_indx = np.zeros((num_eqn[1], 1))
	# total number of Jacobian elements per equation
	njac = np.zeros((num_eqn[1], 1))
	# indices of Jacobian to affect per equation (rows)
	jac_indx = np.zeros((num_eqn[1], 1))
	# ---------------------------------------------------------------------

	max_no_reac = 0.0 # log maximum number of reactants in a reaction
	max_no_prod = 0.0 # log maximum number of products in a reaction

	# loop through aqueous-phase equations line by line and extract the required information
	for eqn_step in range(num_eqn[1]):
		
		line = aqeqn_list[eqn_step] # extract this line
		
		# work out whether equation or reaction rate coefficient part comes first
		eqn_start = str('.*\\' +  chem_scheme_markers[10])
		rrc_start = str('.*\\' +  chem_scheme_markers[9])
		# get index of these markers, note span is the property of the match object that
		# gives the location of the marker
		eqn_start_indx = (re.match(eqn_start, line)).span()[1]
		rrc_start_indx = (re.match(rrc_start, line)).span()[1]
		
		if eqn_start_indx>rrc_start_indx:
			eqn_sec = 1 # equation is second part
		else:
			eqn_sec = 0 # equation is first part
		
		# split the line into 2 parts: equation and rate coefficient
		# . means match with anything except a new line character., when followed by a * 
		# means match zero or more times (so now we match with all characters in the line
		# except for new line characters, so final part is stating the character(s) we 
		# are specifically looking for, \\ ensures the marker is recognised
		if eqn_sec == 1:
			eqn_markers = str('\\' +  chem_scheme_markers[10]+ '.*\\' +  chem_scheme_markers[11])
		else: # end of equation part is start of reaction rate coefficient part
			eqn_markers = str('\\' +  chem_scheme_markers[10]+ '.*\\' +  chem_scheme_markers[9])

		# extract the equation as a string ([0] extracts the equation section and 
		# [1:-1] removes the bounding markers)
		eqn = re.findall(eqn_markers, line)[0][1:-1].strip()
		
		eqn_split = eqn.split()
		eqmark_pos = eqn_split.index('=')
		# with stoich number; rule out the photon
		reactants = [i for i in eqn_split[:eqmark_pos] if i != '+' and i != 'hv']
		products = [t for t in eqn_split[eqmark_pos+1:] if t != '+'] # with stoich number
		
		# record maximum number of reactants across all equations
		max_no_reac = np.maximum(len(reactants), max_no_reac)
		# record maximum number of products across all equations
		max_no_prod = np.maximum(len(products), max_no_prod)

		# append columns if needed
		while max_no_reac > np.minimum(rindx.shape[1], rstoi.shape[1]): 
			rindx = np.append(rindx, (np.zeros((num_eqn[1], 1))).astype(int), axis=1)
			rstoi = np.append(rstoi, (np.ones((num_eqn[1], 1))), axis=1)
			y_arr = np.append(y_arr, (np.ones((num_eqn[1], 1))*-9999).astype(int), axis=1)
			y_arr_fixer = ((np.arange(0, num_eqn[1], dtype = 'int')).reshape(-1, 1)).repeat(max_no_reac, axis=1)
			y_arr[y_arr!=-9999] = y_arr[y_arr!=-9999]+y_arr_fixer[y_arr!=-9999] 
		while max_no_prod > np.minimum(pindx.shape[1], pstoi.shape[1]): 
			pindx = np.append(pindx, (np.zeros((num_eqn[1], 1))).astype(int), axis=1)
			pstoi = np.append(pstoi, (np.zeros((num_eqn[1], 1))), axis=1)
		while ((len(reactants)**2.0+len(reactants)*len(products))>jac_indx.shape[1]):
			jac_indx = np.append(jac_indx, (np.zeros((num_eqn[1], 1))), axis=1)
			jac_den_indx = np.append(jac_den_indx, (np.zeros((num_eqn[1], 1))), axis=1)		
			jac_stoi = np.append(jac_stoi, (np.zeros((num_eqn[1], 1))), axis=1)

		# .* means occurs anywhere in line and, first \ means second \ can be interpreted 
		# and second \ ensures recognition of marker
		rate_coeff_start_mark = str('\\' +  chem_scheme_markers[9])
		# . means match with anything except a new line character, when followed by a * 
		# means match zero or more times (so now we match with all characters in the line
		# except for new line characters, \\ ensures the marker
		# is recognised
		if eqn_sec == 1: # end of reaction rate coefficient part is start of equation part
			rate_coeff_end_mark = str('.*\\' +  chem_scheme_markers[10])
		else: # end of reaction rate coefficient part is end of line
			rate_coeff_end_mark = str('.*\\' +  chem_scheme_markers[11])
		
		# rate coefficient starts and end punctuation
		rate_regex = str(rate_coeff_start_mark + rate_coeff_end_mark)
		# rate coefficient expression in a string
		rate_ex = re.findall(rate_regex, line)[0][1:-1].strip()

		# convert fortran-type scientific notation to python type
		rate_ex = formatting.SN_conversion(rate_ex)
		# convert the rate coefficient expressions into Python readable commands
		rate_ex = formatting.convert_rate_mcm(rate_ex)
		if (rate_ex.find('EXP') != -1):
			print('Error in reaction rate coefficient expression: ', rate_ex)
			sys.exit()
		
		# store the reaction rate coefficient for this equation 
		# (/s once any inputs applied)
		reac_coef.append(rate_ex)
		
		# extract the stoichiometric number of the component in current equation
		reactant_step = 0
		product_step = 0
		stoich_regex = r"^\d*\.\d*|^\d*"
		numr = len(reactants) # number of reactants in this equation
		
		
		# left hand side of equations (losses)
		for reactant in reactants:
				
			if (re.findall(stoich_regex, reactant)[0] != ''):
				stoich_num = float(re.findall(stoich_regex, reactant)[0])
				# name with no stoich number
				name_only = re.sub(stoich_regex, '', reactant)
			elif (re.findall(stoich_regex, reactant)[0] == ''):
				stoich_num = 1.0
				name_only = reactant
			
			# store stoichiometry
			rstoi[eqn_step, reactant_step] = stoich_num
			jac_stoi[eqn_step, reactant_step] = -1*stoich_num

			if name_only not in comp_namelist: # if new component encountered
				comp_namelist.append(name_only) # add to chemical scheme name list
			
				# convert MCM chemical names to SMILES
				if name_only in spec_name:
					# index where xml file name matches reaction component name
					name_indx = spec_name.index(name_only)
					name_SMILE = spec_smil[name_indx] # SMILES of component
				else:
					print(str('Error: inside eqn_parser, chemical scheme name '+str(name_only)+' not found in xml file'))
					sys.exit()
			
				comp_list.append(name_SMILE) # list SMILE names
				name_indx = comp_num # allocate index to this species
				# Generate pybel
				Pybel_object = pybel.readstring('smi', name_SMILE)
				# append to Pybel object list
				Pybel_objects.append(Pybel_object)
				
				comp_num += 1 # number of unique species
				

			else: # if it's a species already encountered it will be in comp_list
				# existing index
				name_indx = comp_namelist.index(name_only)
			
			# store reactant index
			# check if index already present - i.e. component appears more than once
			if sum(rindx[eqn_step, 0:reactant_step]==int(name_indx))>0:
				# get existing index of this component
				exist_indx = (np.where(rindx[eqn_step, 0:reactant_step]==(int(name_indx))))[0]
				# add to existing stoichiometry
				rstoi[eqn_step, exist_indx] += rstoi[eqn_step, reactant_step]
				jac_stoi[eqn_step, exist_indx] += -1*rstoi[eqn_step, reactant_step]
				# remove stoichiometry added above
				rstoi[eqn_step, reactant_step] = 1
				jac_stoi[eqn_step, reactant_step] = 0
				reactant_step -= 1 # ignore this duplicate
			else:
				rindx[eqn_step, reactant_step] = int(name_indx)
				y_arr[eqn_step, reactant_step] = int((eqn_step*max_no_reac)+reactant_step)
				y_rind = np.append(y_rind, int(name_indx))
				rr_arr = np.append(rr_arr, int(eqn_step))
		
			reactant_step += 1
					
		# number of reactants in this equation
		nreac[eqn_step] = int(reactant_step)
		
		# record 1D array of stoichiometries per equation
		rstoi_flat = np.append(rstoi_flat, rstoi[eqn_step, 0:int(reactant_step)])
		
		# right hand side of equations (gains)
		for product in products:

			if (re.findall(stoich_regex, product)[0] != ''):
				stoich_num = float(re.findall(stoich_regex, product)[0])
				name_only = re.sub(stoich_regex, '', product) # name with no stoich number

			elif (re.findall(stoich_regex, product)[0] == ''):
				stoich_num = 1.0
				name_only = product
			
			# store stoichiometry
			pstoi[eqn_step, product_step] = stoich_num
			jac_stoi[eqn_step, reactant_step+product_step] = 1*stoich_num
			if name_only not in comp_namelist: # if new component encountered
				comp_namelist.append(name_only)
				
				# convert MCM chemical names to SMILES
				# index where xml file name matches reaction component name
				if name_only in spec_name:
					name_indx = spec_name.index(name_only)
					name_SMILE = spec_smil[name_indx]
				else:
					print('Error: inside eqn_interr, chemical scheme name '+str(name_only)+' not found in xml file')
					sys.exit()
				
				comp_list.append(name_SMILE) # list SMILE string of parsed species
				name_indx = comp_num # allocate index to this species
				
				# generate pybel object
				Pybel_object = pybel.readstring('smi', name_SMILE)
				# append to Pybel object list
				Pybel_objects.append(Pybel_object)
				
				comp_num += 1 # number of unique species
				
			

			else: # if it's a species already encountered
				# index of component already listed
				name_indx = comp_namelist.index(name_only)
				
			# store product index
			# check if index already present - i.e. component appears more than once
			if sum(pindx[eqn_step, 0:product_step]==int(name_indx))>0:
				# get existing index of this component
				exist_indx = (np.where(pindx[eqn_step, 0:product_step]==(int(name_indx))))[0]
				# add to existing stoichiometry
				pstoi[eqn_step, exist_indx] += pstoi[eqn_step, product_step]
				jac_stoi[eqn_step, reactant_step+exist_indx] += 1*pstoi[eqn_step, product_step]
				# remove stoichiometry added above
				pstoi[eqn_step, product_step] = 0
				jac_stoi[eqn_step, reactant_step+product_step] = 0
				product_step -= 1 # ignore this duplicate
			else:
				pindx[eqn_step, product_step] = int(name_indx)
				rr_arr_p = np.append(rr_arr_p, int(eqn_step))
				y_pind = np.append(y_pind, int(name_indx))

			product_step += 1
		
		# number of products in this equation
		nprod[eqn_step] = int(product_step)
		# record 1D array of stoichiometries per equation
		pstoi_flat = np.append(pstoi_flat, pstoi[eqn_step, 0:int(product_step)])
		
		# now that total number of components (reactants and products) 
		# in an equation is known, repeat the reactant indices over all 
		# components
		tot_comp = nreac[eqn_step]+nprod[eqn_step]
		for i in range(nreac[eqn_step]):
			jac_den_indx[eqn_step, i*tot_comp:(i+1)*tot_comp] = rindx[eqn_step, i]
			# also repeat the stoichiometries for every reactant
			if (i > 0):
				jac_stoi[eqn_step, i*tot_comp:(i+1)*tot_comp] = jac_stoi[eqn_step, 0:tot_comp] 
 		# number of Jacobian elements affected by this equation
		njac[eqn_step, 0] = tot_comp*nreac[eqn_step]
	
	# account for gas-phase in Jacobian denominator index
	jac_den_indx += (comp_num+2)

	# flatten index for arranging concentrations ready for reaction rate coefficient calculation
	y_arr_aq = y_arr.flatten(order='C')
	y_arr_aq = y_arr[y_arr != -9999] # remove fillers
	y_rind_aq = y_rind.astype(int) # ensure integer type
	uni_y_rind_aq = (np.unique(y_rind)).astype(int) # unique index of reactants
	y_pind_aq = y_pind.astype(int) # ensure integer type
	uni_y_pind_aq = (np.unique(y_pind)).astype(int) # unique index of products
	rr_arr_aq = rr_arr.astype(int) # ensure integer type
	rr_arr_p_aq = rr_arr_p.astype(int) # ensure integer type	
	# colptrs for sparse matrix of the change to reactants per equation
	reac_col_aq = np.cumsum(nreac)-nreac
	# colptrs for sparse matrix of the change to products per equation
	prod_col_aq = np.cumsum(nprod)-nprod
	if (len(reac_col_aq) > 0): # if aqueous-phase reaction present	
		# include final columns
		reac_col_aq = np.append(reac_col_aq, reac_col_aq[-1]+nreac[-1])
		prod_col_aq = np.append(prod_col_aq, prod_col_aq[-1]+nprod[-1])
	
	# tag other aqueous-phase arrays
	rindx_aq = rindx
	pindx_aq = pindx
	rstoi_aq = rstoi
	pstoi_aq = pstoi
	jac_stoi_aq = jac_stoi
	rstoi_flat_aq = rstoi_flat
	pstoi_flat_aq = pstoi_flat
	nreac_aq = nreac
	nprod_aq = nprod
	reac_coef_aq = reac_coef
	jac_den_indx_aq = jac_den_indx.astype(int)
	njac_aq = njac.astype(int)
	jac_indx_aq = jac_indx
	jac_indx_aq = jac_indx_aq.astype(int)


	return(rindx_g, rstoi_g, pindx_g, pstoi_g, reac_coef_g, 
			nreac_g, nprod_g, jac_stoi_g, 
			jac_den_indx_g, njac_g, jac_indx_g, 				
			y_arr_g, y_rind_g, uni_y_rind_g, y_pind_g, 
			uni_y_pind_g, reac_col_g, prod_col_g, rstoi_flat_g, pstoi_flat_g, 
			rr_arr_g, rr_arr_p_g, rindx_aq, rstoi_aq, pindx_aq, pstoi_aq, reac_coef_aq, 
			nreac_aq, nprod_aq, jac_stoi_aq, 
			jac_den_indx_aq, njac_aq, jac_indx_aq, 				
			y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
			uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
			rr_arr_aq, rr_arr_p_aq, comp_namelist, comp_list, Pybel_objects, 
			comp_num)



'''modifying the Jacobian inputs per integration time interval to account for changing particle phase fractions'''
# When several components contribute similar fractions to the particle phase of a size bin, the Jacobian 
# appears to work well (relatively low processing time) with only the diagonal elements complete.  However,
# when one or two components dominate the fraction, lower processing times can be achieved by stating
# the row elements of these components in the Jacobian

import numpy as np

def jac_up(Cp, rowvals, colptrs, num_asb, num_comp, jac_part_indx, H2Oi, Cw, jac_wall_indx, ser_H2O): # define function

	# inputs: --------------------------------------------------------------------
	# Cp - particle-phase concentrations (molecules/cc (air))
	# rowvals - rows of Jacobian
	# colptrs - index of rows in rows variable associated with each Jacobian column
	# num_asb - number of particle-phase size bins
	# num_comp - number of components
	# jac_part_indx - the original indices for the sparse Jacobian matrix representing the diagonal
	# 		elements for gas-particle partitioning
	# H2Oi - index for the water component
	# Cw - gas-phase concentration of water (molecules/cc (air))
	# jac_wall_indxn - the original indices for sparse Jacobian matrix representing the diagonal
	#	elements for gas-wall partitioning
	# ser_H2O - whether water gas-particle partitioning serialised
	# ------------------------------------------------------------------------------
	
	# if no size bins or no water or water gas-particle partitioning being solved separately, then nothing to change
	if ((num_asb == 0) or (Cw == 0.) or (ser_H2O == 1)):
		return(rowvals, colptrs, jac_part_indx, 0, [], np.ones((num_asb,1))*-1, jac_wall_indx, [])
		
	# identify rows of Jacobian to be affected - these indices relate to
	# the gas-phase rows of the Jacobian
	rw_indx = (np.array(H2Oi)).repeat(num_asb)
	
	# Number of elements in Jacobian to be modified as a result of high particle-phase 
	# fraction.  Multiply by two to account for both the gas and particle.  Multiply by
	# the number of component less itself since all columns per size bin will be affected
	# except for itself where the diagonal is already accounted for
	jac_mod_len = int(num_asb)*2.*(num_comp-1)
	
	# -------------------------------------------------------------------------------------------------------------
	# number of new Jacobian elements to be included from considering the particle-phase 
	# water effect, multiply by two to include effect on both the gas and particle phase
	jac_part_H2O_eff_num = (num_comp-1)*2*num_asb
	# ------------------------------------------------------------------------------------------------------------
	
	# number of jacobian particle effect indices allocated for gas-on-gas and 
	# gas-on-particle (starting index for jacobian particle effect indices 
	# (particle-on-gas and particle-on-particle effect))
	jpi_st = num_comp*(num_asb+1)
	
	jac_mod_len = 0 # count on the new number of elements in sparse Jacobian
	
	# new array for holding the indices of the sparse jacobian matrix that represent
	# the gas-particle partitioning diagonal elements 
	jac_part_indxn = np.zeros((len(jac_part_indx)))
	jac_part_indxn[:] = (jac_part_indx[:])
	jac_part_indxn = jac_part_indxn.astype('int')
	# same for gas-wall partitioning
	jac_wall_indxn = np.zeros((len(jac_wall_indx)))
	jac_wall_indxn[:] = (jac_wall_indx[:])
	jac_wall_indxn = jac_wall_indxn.astype('int')
	
	# new array for holding row indices of the sparse Jacobian
	rowvalsn = np.zeros((len(rowvals)))
	rowvalsn[:] = rowvals[:]
	rowvalsn = rowvalsn.astype('int')
	
	# new array for holding column indices of the sparse Jacobian
	colptrsn = np.zeros((len(colptrs)))
	colptrsn[:] = colptrs[:]
	colptrsn = colptrsn.astype('int')
	
	# array to hold sparse Jacobian indices representing the particle-on-gas and particle-on-particle
	# effect on particle-phase water
	jac_part_hmf_indx = np.zeros((0), dtype = int)
	
	# array to hold sparse Jacobian indices representing the particle-on-gas and particle-on-particle
	# effect of particle-phase water
	jac_part_H2O_indx = np.zeros((0), dtype = int)
	
	
	# modify Jacobian sparse matrix inputs to account for water in the particle phase
	for isb in range(int(num_asb)): # loop through size bins
			
		# starting indices for diagonal matrix for this size bin: first term described above, 
		# second is default inputs due to particle-on-gas and particle-on-particle diagonals 
		pi_sti = int(jpi_st+isb*(num_comp*2))
		# finishing indices for diagonal matrix for this size bin: terms described above
		pi_fii = int(jpi_st+(isb+1)*(num_comp*2))
			
		# new number of elements for Jacobian, multiply by 2 because of particle
		# effect on gas and particle, subtract 1 because diagonal already 
		# accounted for
		new_num_ele = (num_comp-1)*2
			
		# length of increment array to be applied to jacobian particle-phase diagonal part,
		# multiply by two to account for the gas and particle part
		inc_len = int(num_comp*2)
			
		# prepare array holding the increment per diagonal element
		indx_incr = np.arange(inc_len)
			
		# identify where diagonal occurs as no additional jacobian elements needed here
		comp_indx = rw_indx[isb]
		
		# the increments needed for the jacobian particle index
		indx_incr[comp_indx*2+1::] -= 1
			
		# repeat final increment over any larger size bins
		indx_incr = (np.append(indx_incr, np.ones(((num_asb-(isb+1))*num_comp*2))*indx_incr[-1])).astype('int')
			
		# new indices for jacobian particle-on-gas and particle-on-particle diagonals
		jac_part_indxn = np.concatenate((jac_part_indxn[0:pi_sti], (jac_part_indxn[pi_sti::]+indx_incr)))
		# push up the wall-on-gas and gas-on-wall elements of the gas-wall partitioning indices
		jac_wall_indxn[num_comp*2::] += (num_comp-1)*2
			
		# Jacobian indices for water's particle-on-gas and particle-on-particle rows
		ji_indx = ((jac_part_indxn[pi_sti+1:pi_fii])-(jac_part_indxn[pi_sti:pi_fii-1]) == 2)
			
		# new indices for the jacobian particle-on-gas and particle-on-particle
		# rows for water
		jac_part_hmf_indx = np.concatenate((jac_part_hmf_indx, ((jac_part_indxn[pi_sti+1:pi_fii][ji_indx])-1)))
			
		jac_mod_len += new_num_ele # count on the new number of elements
		# starting rowvals index for this size bin
		rv_indx_st = colptrsn[int(num_comp*(isb+1))]
		# finishing rowvals index for this size bin
		rv_indx_fi = colptrsn[int(num_comp*(isb+2))]
			
		rowvals_pre = rowvalsn[0:rv_indx_st]
		rowvals_pos = rowvalsn[rv_indx_fi::]
		rowvals_st = rowvalsn[rv_indx_st:rv_indx_fi]
			
		# reshape rowvals for this size bin so that components are in columns, effect on gas is in first row and
		# effect on particle is in third row
		rowvals_st = np.transpose(rowvals_st.reshape(int(num_comp), 2))
		rowvals_st = np.concatenate((rowvals_st[0, :].reshape(1, -1), np.ones((1, num_comp))*-1, rowvals_st[1, :].reshape(1, -1), np.ones((1, num_comp))*-1), axis=0)
			
		# shift down the necessary gas effect elements
		rowvals_st[1, rowvals_st[0, :]>comp_indx] = rowvals_st[0, rowvals_st[0, :]>comp_indx]
		rowvals_st[1, rowvals_st[0, :]<comp_indx] = comp_indx
		rowvals_st[0, rowvals_st[1, :]>comp_indx] = comp_indx
		# shift down the necessary particle effect elements
		rowvals_st[3, rowvals_st[2, :]>(comp_indx+(num_comp*(isb+1)))] = rowvals_st[2, rowvals_st[2, :]>(comp_indx+(num_comp*(isb+1)))]
		rowvals_st[3, rowvals_st[2, :]<(comp_indx+(num_comp*(isb+1)))] = (comp_indx+(num_comp*(isb+1)))
		rowvals_st[2, rowvals_st[3, :]>(comp_indx+(num_comp*(isb+1)))] = (comp_indx+(num_comp*(isb+1)))
			
		# ensure still integer
		rowvals_st = rowvals_st.astype('int')
			
		# ensure values to be used as indices are integer
		num_comp = int(num_comp)
			
		# get the new number of row indices per component, subtract the number prior to modification
		col_incr = np.sum(rowvals_st != -1, axis=0)-(colptrs[(num_comp*(isb+1))+1:(num_comp*(isb+2)+1)]-colptrs[(num_comp*(isb+1)):(num_comp*(isb+2))])
			
		# account for new Jacobian elements in colptrs
		colptrsn[(num_comp*(isb+1)+1):(num_comp*(isb+2)+1)] += np.cumsum(col_incr)
		# also account for in Jacobian elements further on
		colptrsn[(num_comp*(isb+2)+1)::] += (np.cumsum(col_incr))[-1]
			
		# flatten new rowvals into 1D array
		rowvals_st = rowvals_st.flatten(order = 'F')
		# keep only wanted elements
		rowvals_st = rowvals_st[rowvals_st != -1]
		# concatenate back to original
		rowvalsn = np.concatenate((rowvals_pre, rowvals_st, rowvals_pos))
		
		# ------------------------------------------------------------------------------------------------
		# include effect of particle-phase water on gas and particle phase of other components
		# (the particle-phase water column in the Jacobian)
		
		# first element of the particle effect diagonal indices array to be affected, first term is the 
		# elements affected by the gas phase, second term are elements affected in smaller size 
		# bin particle phase and final term is the relevant index in this size bin
		#jpi_indx = num_comp*(num_asb+1)+(num_comp*2*isb)+H2Oi*2
		# starting index for sparse Jacobian
		#jac_part_H2O_sti = jac_part_indxn[jpi_indx]
		
		# particle-on-gas
		# new indices before diagonal
		#ni = np.arange(jac_part_H2O_sti, jac_part_H2O_sti+H2Oi)
		#jac_part_H2O_indx = np.concatenate((jac_part_H2O_indx, ni))
		# push up the indices for the particle diagonal indices
		#jac_part_indxn[jpi_indx::] += len(ni)
		# push up the indices for the particle row indices
		#jac_part_hmf_indx[H2Oi*2+(((num_comp-1)*2)*isb)::] += len(ni)
		
		# new indices after diagonal
		#ni = np.arange(jac_part_H2O_sti+H2Oi+1, (jac_part_H2O_sti+H2Oi+1)+((num_comp-1)-H2Oi))
		#jac_part_H2O_indx = np.concatenate((jac_part_H2O_indx, ni))
		# push up the indices for the particle diagonal indices
		#jac_part_indxn[(jpi_indx+1)::] += len(ni)
		# push up the indices for the particle row indices
		#jac_part_hmf_indx[H2Oi*2+(((num_comp-1)*2)*isb)::] += len(ni)
		
		# push up the wall-on-gas and gas-on-wall elements of the gas-wall partitioning indices
		#jac_wall_indxn[num_comp*2::] += (num_comp-1)
		
		# ---------------------------------------
		# particle-on-particle
		# second element of the particle effect diagonal indices array to be affected
		#jpi_indx = num_comp*(num_asb+1)+(num_comp*2*isb)+H2Oi*2+1
		# starting index of sparse Jacobian
		#jac_part_H2O_sti = jac_part_indxn[jpi_indx]
		# new indices before diagonal
		#ni = np.arange(jac_part_H2O_sti, jac_part_H2O_sti+H2Oi)
		#jac_part_H2O_indx = np.concatenate((jac_part_H2O_indx, ni))
		
		# push up the indices for the particle diagonal indices
		#jac_part_indxn[jpi_indx::] += len(ni)
		# push up the indices for the particle row indices
		#jac_part_hmf_indx[H2Oi*2+(((num_comp-1)*2)*isb)::] += len(ni)
		
		# new indices after diagonal
		#ni = np.arange(jac_part_H2O_sti+H2Oi+1, (jac_part_H2O_sti+H2Oi+1)+((num_comp-1)-H2Oi))
		#jac_part_H2O_indx = np.concatenate((jac_part_H2O_indx, ni))
		# push up the indices for the particle diagonal indices
		#jac_part_indxn[(jpi_indx+1)::] += len(ni)
		# push up the indices for the particle row indices
		#jac_part_hmf_indx[H2Oi*2+(((num_comp-1)*2)*isb)::] += len(ni)
		
		# push up the wall-on-gas and gas-on-wall elements of the gas-wall partitioning indices
		#jac_wall_indxn[num_comp*2::] += (num_comp-1)
		
		# -----------------------------------------------------
		# increment rowvals and colptrs in response to new sparse Jacobian points
		#rowvals_pre = rowvalsn[0:colptrsn[num_comp+(num_comp*isb)+H2Oi]]
		#rowvals_pro = rowvalsn[colptrsn[num_comp+(num_comp*isb)+H2Oi+1]::]
		# note that with the two segments of the original rowvals above (rowvals_pre & rowvals_pro) 
		# the diagonal is omitted so need to account for this in the new rowals
		#rowvals_new = np.concatenate((np.arange(0, num_comp), np.arange(num_comp*(isb+1), num_comp*(isb+2)))) 
		#rowvalsn = np.concatenate((rowvals_pre, rowvals_new, rowvals_pro))
		#colptrsn[(num_comp+(num_comp*isb)+H2Oi+1)::] += (num_comp-1)*2
		 
		#jac_mod_len += (num_comp-1)*2 # count on the new number of elements in sparse Jacobian
	
	return(rowvalsn, colptrsn, jac_part_indxn, jac_mod_len, jac_part_hmf_indx, rw_indx, jac_wall_indxn, jac_part_H2O_indx)


'''determines number of photochemical reactions'''
# based on the photochemistry file path, the number of reactions is
# inferred

import os

# define function
def photo_num(photo_file):

	# inputs: -------------------------------------------------------------
	# photo_file - path to file containing abosrption cross-sections and
	# quantum yields
	# ---------------------------------------------------------------------
	# number of photolysis reactions, if this relevant
	cwd = os.getcwd() # address of current working directory
	if photo_file == str(cwd + '/PyCHAM/photofiles/MCMv3.2'):
		Jlen = 62 # for MCM (default name of photolysis parameters)
	else: # need to find out number of photolysis reactions
		# use Fortran indexing to be consistent with MCM photochemical reaction numbers
		Jlen = 1 
		# open file to read
		f = open(str(photo_file), 'r')
		for line in f: # loop through line
			if line.strip() == str('J_'+str(Jlen) + '_axs'):
				Jlen += 1
	return(Jlen)



'''function to initiate concentrations of components'''
# based on inputs, initial concentrations and their holding arrays 
# are set

import numpy as np
import scipy.constants as si
import math
from water_calc import water_calc
import write_dydt_rec
import pybel


def init_conc(num_comp, Comp0, init_conc, TEMP, RH, PInit, Pybel_objects,
	testf, pconc, dydt_trak, end_sim_time, save_step, 
	rindx, pindx, num_eqn, nreac, nprod, 
	comp_namelist, Compt, seed_name, seed_mw,
	core_diss, nuc_comp, comp_xmlname, comp_smil):
		
	# inputs:------------------------------------------------------
	
	# num_comp - number of unique components
	# Comp0 - chemical scheme names of components present at start of experiment
	# init_conc - initial concentrations of components (molecules/cc)	
	# TEMP - temperature in chamber at start of experiment (K)
	# RH - relative humidity in chamber (dimensionless fraction 0-1)
	# PInit - initial pressure (Pa)
	# init_SMIL - SMILES of components present at start of experiment (whose 
	# concentrations are given in init_conc)
	# testf - flag for whether in normal mode (0) or testing mode (1/2)
	# pconc - initial concentration of particles (# particles/cc (air))
	# dydt_trak - chemical scheme name of components for which user wants the tendency to  
	#			change tracked
	# end_sim_time - total simulation time (s)
	# save_step - recording frequency (s)
	# num_eqn - number of equations
	# comp_namelist - list of components' names in chemical equation file
	# Compt - name of component injected after start of experiment
	# seed_name - name of core component (input by user)
	# seed_mw - molecular weight of seed material (g/mol)
	# core_diss - dissociation constant of seed material
	# nuc_comp - name of nucleating component (input by user, or defaults to 'core')
	# comp_xmlname - component names in xml file
	# comp_smil - SMILE strings in xml file
	# -----------------------------------------------------------

	if testf==1: # testing mode
		# return dummies
		return(0,0,0,0,0,0,0,0)

	NA = si.Avogadro # Avogadro's number (molecules/mol)
	# empty array for storing species' concentrations, must be an array
	y = np.zeros((num_comp))
	y_mw = np.zeros((num_comp, 1)) # species' molecular weight (g/mol)
	# empty array for storing index of interesting gas-phase components
	y_indx_plot = []
	
	# convert concentrations
	# total number of molecules in 1 cc air using ideal gas law.  R has units cc.Pa/K.mol
	ntot = PInit*(NA/(8.3144598e6*TEMP))
	# one billionth of number of molecules in chamber unit volume
	Cfactor = ntot*1.0e-9 # ppb-to-molecules/cc
	

	# prepare dictionary for tracking tendency to change of user-specified components
	dydt_vst = {}

	# insert initial concentrations where appropriate
	for i in range(len(Comp0)):
    		# index of where initial components occur in list of components
		try: # in case components already listed via interpretation of the chemical scheme
			y_indx = comp_namelist.index(Comp0[i])
			
		# if component not already listed via interpretation of the chemical scheme
		# then add to list
		except:
			print('Note: component specified in initial gas-phase concentrations of model variables has not been registered in the chemical scheme, but will try to be added to the component list inside the module for initialising gas-phase concentrations')
			comp_namelist.append(Comp0[i]) # add to name list
			
			# convert MCM chemical names to SMILES
			if (Comp0[i] in comp_xmlname):
				# index where xml file name matches reaction component name
				name_indx = comp_xmlname.index(Comp0[i])
				name_SMILE = comp_smil[name_indx] # SMILES of component
				# Generate pybel
				Pybel_object = pybel.readstring('smi', name_SMILE)
				# append to Pybel object list
				Pybel_objects.append(Pybel_object)
				
				comp_num += 1 # number of unique species
				
				y_indx = len(comp_namelist)-1
				y = y.append(np.zeros((1)))
			else:
				print(str('Error: inside eqn_parser, chemical scheme name '+str(name_only)+' not found in xml file'))
				sys.exit()
			
		y[y_indx] = init_conc[i]*Cfactor # convert from ppb to molecules/cc (air)
		# remember index for plotting gas-phase concentrations later
		y_indx_plot.append(y_indx)

	# number of recording steps
	nrec_steps = int(math.ceil(end_sim_time/save_step)+1)

	# get index of user-specified components for tracking their dydt due to model 
	# mechanisms
	if len(dydt_trak)>0:
		
		dydt_traki = [] # empty list for indices of these components
		
		for i in range (len(dydt_trak)):
			reac_index = [] # indices of reactions involving this species
			# index of where initial species occurs in SMILE string
			y_indx = comp_namelist.index(dydt_trak[i])

			# remember index for plotting gas-phase concentrations later
			dydt_traki.append(int(y_indx))
			# search through reactions to see where this component is reactant or product
			for ri in range(num_eqn):
				if sum(rindx[ri,0:nreac[ri]]==y_indx)>0:
					reac_index.append(int(ri)) # append reaction index
			for ri in range(num_eqn): # repeat for products
				if sum(pindx[ri,0:nprod[ri]]==y_indx)>0:
					reac_index.append(int(ri)) # append reaction index
					
	
			# save reaction indices in dictionary value for this component,
			# when creating empty rec_array, add two rows onto the end for particle- and 
			# wall-partitioning, respectively.  Note the extra row to hold the reacion 
			# indices
			rec_array = np.zeros((nrec_steps+1, len(reac_index)+2))
			rec_array[0, 0:-2] = (reac_index)
			dydt_vst[y_indx] = rec_array # dictionary entry to hold results

		dydt_vst['comp_index'] = dydt_traki # dictionary entry to hold component names
		
		# call on write_dydt_rec to generate the module that will process
		# the tendency to change during the simulation
		write_dydt_rec.write_dydt_rec()
	
	for i in range(num_comp): # loop through all components
		y_mw[i] = Pybel_objects[i].molwt # molecular weight (g/mol)
	
	# ------------------------------------------------------------------------------------
	# account for water's properties
	H2Oi = num_comp # index for water
	num_comp += 1 # update number of species to account for water
	
	# update gas-phase concentration (molecules/cc (air)) and vapour pressure
	# of water (log10(atm))
	[C_H2O, Psat_water, H2O_mw] = water_calc(TEMP, RH, si.N_A)
	
	# append empty element to y and y_mw to hold water values
	y = np.append(y, C_H2O)
	y_mw = (np.append(y_mw, H2O_mw)).reshape(-1, 1)
	comp_namelist.append('H2O') # append water's name to component name list

	# ------------------------------------------------------------------------------------
	# account for seed properties - note that even if no seed particle, this code ensures
	# that an index is provided for core material

	# empty array for index of core component
	seedi = (np.zeros((len(seed_name)))).astype(int)
	
	comp_namelist.append('core') # append name of core to component name list
	corei = [num_comp] # index for core component
	# increase number of components to account for 'core' component
	num_comp += 1

	# append core gas-phase concentration (molecules/cc (air)) and molecular 
	# weight (g/mol) (needs to have a 1 length in second dimension for the kimt 
	# calculations)
	y = np.append(y, 0.)
	y_mw = (np.append(y_mw, seed_mw)).reshape(-1, 1)
		
	# if nucleating component formed of core component
	if (nuc_comp[0] == 'core'):
		nuci = num_comp-1 # index of core component
	else:
		nuci = -1 # filler
	
	# get indicices of seed particle component(s)
	indx = 0 # count on seed component(s)
	for sname in seed_name:
		# index of core component
		seedi[indx] = int(comp_namelist.index(sname))
		indx += 1 # count on seed component(s)
	
	# get index of component with latter injections
	if len(Compt)>0:
		inj_indx = np.zeros((len(Compt)))
		for i in range(len(Compt)):
			# index of where initial species occurs in SMILE string
			inj_indx[i] = comp_namelist.index(Compt[i])
	else:
		inj_indx = np.zeros((1)) # dummy

	# ensure index arrays are integer type
	inj_indx = inj_indx.astype('int')
	corei = np.array((corei)).astype('int')
	
	return (y, H2Oi, y_mw, num_comp, Cfactor, y_indx_plot, corei, dydt_vst, 
				comp_namelist, inj_indx, core_diss,
				Psat_water, nuci, nrec_steps, seedi)



'''module to check whether ode solver estimates partitioning that leads to unacceptable change in volume of particles'''
# code called by the mov_cen_main.py module to ensure that the volume change of particles
# estimated by the ode solver is within acceptable bounds

import numpy as np
from compl_evap import compl_evap as compl_evap

def Vchange_check(res, MV, sbb, sbn, NA, n0, nc, solv_time, ts0, bc_red, Vol0, Psat):

	# inputs: ---------------------------------
	
	# res - concentrations (molecules/cc (air)) of all components in all phases (columns) 
	#		at the adaptive time steps set by the ode solver
	# MV - molar volume of components (cc/mol)
	# sbb - fixed volume bounds (um3) of size bins
	# sbn - number of size bins
	# NA - Avogadro's number (molecules/mol)
	# n0 - number concentration of particles per size bin (# particle/cc (air)) before 
	#		integration
	# nc - number of components
	# solv_time - adaptive time step (s) used by ode solver
	# ts0 - original time step passed to integrator (s)
	# bc_red - flag for whether boundary conditions have caused a change to time step
	# Vol0 - default volume per size bin (um3), volume at centre of size bin bounds
	# Psat - saturation vapour pressure of components (molecules/cm3 (air))
	# -----------------------------------------
	
	# flag for breaking condition of volume change, updated once volumes checked below
	Vchang_flag = 2
	# count on adaptive time steps - work backwards through the adaptive time steps of the
	# ode solver to check whether volume condition met
	tsi = 1
	# acceptable number of size bins to change by
	acc_sb_chng = 1000
	Vnew = np.zeros((sbn))
	
	while Vchang_flag>=1:
		
		# estimated concentrations (moleculecs/cc (air)) at this time step
		ytest = res
		
		# loop through size bins to check whether volume condition met
		for sbi in range(sbn):
			
			# calculating new volumes of particles in size bins --------------------------
			if n0[sbi]<1.0e-10: # if no particles, assign the default volume (um3)
				Vnew[sbi] = Vol0[sbi]
				continue # onto next size bin
			# concentration of components in this size bin (molecules/cc (air))
			Cnow = ytest[(sbi+1)*nc:(sbi+2)*nc]
			# molar concentration of components in one particle (mol/cc (air))
			Cnow = Cnow/(NA*n0[sbi])
			# new volume (um3) of single particles, note MV has units cc/mol, so needs 
			# conversion to um3/mol
			Vnew[sbi] = sum(Cnow*(MV[:, 0]*1.e12))
			
			# comparing new volumes against condition for volume change ------------------
			# check for volume conditions 
			if (Vnew[sbi]<0.0): # shrunk to unrealistic negative volume
				# allow complete evaporation of smallest size bin if only volatiles 
				# present and time step already relatively small
				if sbi == 0.0 and sum(Cnow[Psat<1.0e-20]<1.0e-20) and ts0<1.0e-3:
					(ytest, n0, Vnew) = compl_evap(ytest, n0, Vnew, Vol0, nc, sbn)
				else: # change integration time step
					Vchang_flag = 1
# 					print('negative volume seen in Vchange_check.py, size bin affected: ' + str(sbi) + ', concentration of components: ' + str(Cnow))
			if (Vnew[sbi]>sbb[-1]): # grown to unpractical positive volume
				Vchang_flag = 1
# 				print('volume above uppermost size bin bound seen in Vchange_check.py, size bin affected: ' + str(sbi) + ', concentration of components: ' + str(Cnow))
			if ((sbi-acc_sb_chng) >= 1):
				if (Vnew[sbi]<sbb[sbi-acc_sb_chng]): # excessive shrink
					Vchang_flag = 1
					print('volume below acceptable change in Vchange_check.py, size bin affected: ' + str(sbi) + ', concentration of components: ' + str(Cnow) + ', percentage away from allowed change bound: ' + str(((sbb[sbi-acc_sb_chng]-Vnew[sbi])/sbb[sbi-acc_sb_chng])*100.0))
			if (((sbi+1)+acc_sb_chng) < sbn+1):
				if (Vnew[sbi]>sbb[(sbi+1)+acc_sb_chng]): # excessive growth
					Vchang_flag = 1
					print('volume above acceptable change in Vchange_check.py, size bin affected: ' + str(sbi) + ', concentration of components: ' + str(Cnow) + ', percentage away from allowed change bound: ' + str(((Vnew[sbi]-sbb[(sbi+1)+acc_sb_chng])/sbb[(sbi+1)+acc_sb_chng])*100.0))
					
			# decision based on volume change condition ----------------------------------
			if (Vchang_flag==1): # if volume condition not met
				tsi += 1
				# break out of size bin loop and change adaptive time step index
				if (tsi < res.shape[0]):
					Vchang_flag = 3 # update flag and try next integrator time step
# 					print('trying earlier ode solver result, number of results from the final: ', tsi)
					break # stop size bin loop
				else: # need to reduce integration time step (s) below the minimum here
					redt = 1 # flag for time step reduction due to volume change
					t = solv_time[0]/2.0
					bc_red = 0 # flag for time step reduction due to boundary conditions
					Vchang_flag = 0 # will exit while loop for volume condition not met
					break # stop size bin loop
			else: # if volume condition met
				Vchang_flag = 2
		
		# applies if size bin loop finished and volume condition met ---------------------		
		# if volume condition met after all size bins checked adopt the results at this 
		# time
		if (Vchang_flag==2):
			
			if (tsi==1): # integration time step unchanged
				t = ts0 # time for integration on next step (s)
				redt = 0 # flag for no time step reduction due to volume change
			if (tsi>1): # if a sub-step used, then get associated time
				t = solv_time[-tsi] # new time for integration on next step (s)
				redt = 2 # flag for time step reduction due to volume change
			Vchang_flag = 0 # will exit while loop for volume condition not met
			
	return(redt, t, bc_red, Vnew, tsi)



'''module called to run the PyCHAM model, produces graphical user interface'''
from PyQt5.QtWidgets import QApplication, QWidget
import sys
import os
# to push to Github repo:
# switch to wanted branch: git checkout -b branch_name
# if not already linked to github: git init
# followed by: git remote add origin https://github.com/simonom/PyCHAM.git
# then use command above to switch to wanted branch
# git add .
# git commit -m "commit comment"
# git push origin master

def main(): 

	import PyCHAM
	import signal
	
	# what to do if user breaks in terminal using ctrl+c and the 
	# programme is running, specifically when it is not waiting for
	# a button on the GUI to be pressed
	def sig_han(signum, frame):
		sys.exit()
	
	app = QApplication(sys.argv)
	signal.signal(signal.SIGINT, sig_han) # in case of ctrl+c	
	ex = PyCHAM.PyCHAM()
	app.exec_()
	
	return()
	

if __name__ == '__main__':
	main() # call function
	
sys.exit()



'''module to state PyCHAM package metadata'''
from setuptools import setup
from setuptools import find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="PyCHAM",
    version="2.1.9",
    author="Simon O'Meara",
    author_email="simon.omeara@manchester.ac.uk",
    description="PyCHAM: CHemistry with Aerosol Microphysics in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['aerosol', 'chamber', 'model'],
    url="https://github.com/simonom/PyCHAM",
    packages=find_packages(include=['PyCHAM', 'PyCHAM.*']),
    include_package_data=True,
    package_data={'PyCHAM': ['input/*', 'output/example_scheme/Example_Run_output/*', 'output/example_scheme_inc_aq/Example_Run_output_inc_aq/*', 'output/GMD_paper_plotting_scripts/*', 'output/plotting_scripts/*', 'photofiles/*', 'photofiles/MCMv3.2/*', 'photofiles/MCMv3.2/(CH3)3CONO2/*', 'photofiles/MCMv3.2/BIACET/*', 'photofiles/MCMv3.2/C2H5CHO/*', 'photofiles/MCMv3.2/C5HPALD1/*', 'photofiles/MCMv3.2/CH3CH2ONO2/*', 'photofiles/MCMv3.2/CH3CHO/*', 'photofiles/MCMv3.2/CH3COCH3/*', 'photofiles/MCMv3.2/(CH3COCHO/*', 'photofiles/MCMv3.2/CH3ONO2/*', 'photofiles/MCMv3.2/CH3OOH/*', 'photofiles/MCMv3.2/CHOCHO/*', 'photofiles/MCMv3.2/H2O2/*', 'photofiles/MCMv3.2/HCHO/*', 'photofiles/MCMv3.2/HNO3/*', 'photofiles/MCMv3.2/HONO/*', 'photofiles/MCMv3.2/i_C3H7CHO/*', 'photofiles/MCMv3.2/MACR/*', 'photofiles/MCMv3.2/MEK/*', 'photofiles/MCMv3.2/MVK/*', 'photofiles/MCMv3.2/n_C3H7CHO/*', 'photofiles/MCMv3.2/NO2/*', 'photofiles/MCMv3.2/NO3/*', 'photofiles/MCMv3.2/NOA/*', 'photofiles/MCMv3.2/O3/*', 'unit_tests/*', 'unit_tests/input/*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    
# state dependencies for PyCHAM, note that openbabel is installed separately, as described in README
install_requires=["PyQt5","pickleshare","numpy","backcall","certifi","cycler","decorator","gitpython","ipdb","matplotlib","scipy","xmltodict","requests"]
)



'''module to track particle number size distribution using moving centre size structure (p. 416 of Jacobson 2000)'''
# note, no limits on number of size bins a particle can move in
# this version of moving centre

import numpy as np
import scipy.constants as si

def mov_cen_main(n0, Vbou, Cn, sbn, nc, Vol0, t, tinc_count, MV):



	# input:---------------------------------------------------------
	# n0 - particle number concentration per size bin before time step
	# (# particle/cc (air)) (excluding wall)
	# Vbou - volume bounds per size bin (um3) (number of size bins +1)
	# Cn - particle phase concentration per component per size bin
	# (molecules/cc (air)), with rows representing
	# components and columns representing size bins
	# sbn - number of size bins excluding wall (if present)
	# nc - number of components
	# Vol0 - original volume of size bins (um3) (excluding wall)
	# t - integration time (s)
	# tinc_count - count on number of time steps since time interval last required 
	# decreasing
	# MV - molar volume per component (um3/mol)
	# ---------------------------------------------------------------
	
	NA = si.Avogadro # Avogadro's number (molecules/mol)	

	Vnew = np.zeros((sbn))
	ish = n0[:, 0] > 0. # size bins containing particle number concentration
	nmolC = np.zeros((nc, sbn))
	
	# number of moles of each component in a single particle (mol/cc (air))
	nmolC[:, ish] = ((Cn[:, ish]/(NA*n0[ish, 0])))
	MVrep = np.repeat(MV, nmolC.shape[1], axis=1)
	
	
	# new volume of single particle per size bin (um3) including volume of water
	Vnew[:] = np.sum(nmolC*MV, axis=0)
	Vnew[n0[:, 0] <= 1.e-10] = Vol0[0::][n0[:, 0] <= 1.e-10]

	# array of new particle number concentration (# particle/cc (air))
	num_part_new = np.zeros((sbn, 1))
	# array of new molecular concentration (# molecules/cc (air))
	num_molec_new = np.zeros((nc, sbn))
	# loop through size bins and reallocate particle number concentration 
	# (# particle/cc (air)) and molecular concentration (# molecules/cc (air))
	for i in range(sbn):
	
		# return and send message if beyond uppermost boundary
		if Vnew[i]>=Vbou[-1]:
			print('particle in size bin ' +str(i) + ' exceed uppermost volume bound, therefore will try reducing ode solver time step')
			redt = 1
			tnew = t/2.0 # new time for integration on next step (s)
			# note, that when called from water initiator or coag, this condition will
			# terminate the model, so the returned values are meaningless
			return(0, 0, 0, 0, 0, 0, 0)
			
		Vindx = (Vnew[i] >= Vbou[0:-1]).sum()-1 # index of where these particle fit
		num_part_new[Vindx, 0] += n0[i] # allocate number concentration
		num_molec_new[:, Vindx] += Cn[:, i] # allocate molecular concentration
		
	# remove any particles and their constituents if number concentration is relatively very low
	rm_indx = num_part_new[:, 0] < (sum(num_part_new)*1.e-8)
	num_part_new[rm_indx, 0] = 0.
	num_molec_new[:, rm_indx] = 0.
	
	
	# need to find new volumes of single particles (um3) because possible that particles
	# originally in separate size bins now share a size bin
	# total volume of components 
	# ((um3 (all particles)/cc (air))/(particle number/cc (air))) 
	# calculation is:
	# divide number of molecules/cc (air) by Na to get moles/cc(air), then 
	# multiply by um3/mol (MV) to get um3 (of each component)/cc (air),
	# then sum volume of components per size bin to get um3 (all particles)/cc (air)
	MVrep = np.repeat(MV, num_molec_new.shape[1], axis=1)
	
	Vtot = (np.sum((num_molec_new/NA)*(MVrep), 0)) # (um3)

	# then divide by particle number (#/cc (air)) to get volume of single particles
	# per size bin (um3)
	Vsing = np.zeros(len(Vtot))
	isb = num_part_new[:, 0]>0.
	Vsing[isb] = Vtot[isb]/num_part_new[isb, 0]
	
	# new radius of single particles per size bin (um)
	rad = ((3.0*Vsing)/(4.0*np.pi))**(1.0/3.0)
	
	isb = num_part_new[:, 0] == 0.
	
	# fill volume array elements for bins without particles with central volume (um3)
	Vsing[isb] = Vol0[isb]
	rad[isb] = ((3.*Vsing[isb])/(4.*np.pi))**(1./3.) # new radius (um)
	
	# flag to show no reduction in time step needed
	redt = 0
	tnew = t # maintain same time step (s)

	return(num_part_new, Vsing, np.ravel(np.transpose(num_molec_new)), rad, 
			redt, t, tnew)



'''update the chamber variables depending on time'''
# based on the user inputs and time through simulation,
# chamber variables are updated before calling the 
# ODE solver

import numpy as np
import os
import pp_dursim
import volat_calc

# define function
def cham_up(sumt, temp, tempt, Pnow, light_stat, light_time, 
	light_time_cnt, light_ad, tnew, nuc_ad, nucv1, nucv2, nucv3, 
	new_part_sum1, update_stp, update_count, lat, lon, dayOfYear,
	photo_par_file, act_flux_path, injectt, gasinj_cnt, inj_indx, 
	Ct, pmode, pconc, pconct, seedt_cnt, num_comp, y, N_perbin, 
	mean_rad, corei, seedVr, seed_name, lowsize, uppsize, num_sb, MV, rad0, radn, std, 
	y_dens, H2Oi, rbou, const_infl_t, infx_cnt, Cinfl, wall_on, Cfactor, seedi, diff_vol, DStar_org):

	# inputs: ------------------------------------------------
	# sumt - cumulative time through simulation (s)
	# temp - temperature in chamber (K)
	# tempt - times that temperatures reached (s)
	# Pnow - pressure in chamber (Pa)
	# light_stat - status of lights
	# light_time - times that light attain status (s)
	# light_time_cnt - light status counter
	# light_ad - marker for whether to change time interval 
	#	in response to changing natural light intensity
	# tnew - time interval between chamber updates (s)
	# nuc_ad - flag for whether user wants time step adapted 
	# to nucleation
	# nucv1 - nucleation parameter one
	# nucv2 - nucleation parameter two
	# nucv3 - nucleation parameter three
	# new_part_sum1 - total number concentration of new 
	#	particles so far (#/cc (air))
	# update_stp - time interval between operator-split 
	#	updates (s)
	# update_count - count since operator-split last 
	#	updated (s)
	# lat - latitude (degrees)
	# lon - longitude (degrees)
	# dayOfYear - number of days since 31st December
	# photo_par_file - photochemistry parameter file
	# act_flux_path - actinic flux file
	# injectt - time of instantaneous injections of 
	#	components (s)
	# gasinj_cnt - count on injection times of component(s)
	# inj_indx - index of component(s) being injected after 
	#	experiment start
	# Ct - concentration(s) (ppb) of component(s) injected 
	#	instantaneously after experiment start
	# pmode - whether particle number size distributions stated explicitly or by mode
	# pconc - concentration of injected 
	#	particles (#/cc (air))
	# pconct - times of particle injection (s)
	# seedt_cnt - count on injections of particles
	# num_comp - number of components
	# y - concentration of components (molecules/cc (air))
	# N_perbin - concentration of particles (#/cc (air))
	# mean_rad - mean radius for particle number size 
	#	distribution (um)
	# corei - index of core component
	# seedVr - volume ratio of component(s) comprising seed particles
	# seed_name - name(s) of component(s) comprising seed 
	#	particles
	# lowsize - lower size bin boundary (um)
	# uppsize - upper size bin boundary (um)
	# num_sb - number of size bins
	# MV - molar volume of components (cc/mol)
	# rad0 - initial radius at size bin centres (um)
	# radn - current radius at size bin centres (um)
	# std - standard deviation for injected particle number size 
	#	distributions
	# y_dens - component densities (g/cm3)
	# H2Oi - index of water
	# rbou - size bin radius bounds (um)
	# const_infl_t - times for constant influxes (s)
	# infx_cnt - count on constant influx occurrences
	# Cinfl - influx rate for components with constant 
	# 	influx (ppb/s)
	# wall_on - marker for whether wall is on
	# Cfactor - conversion factor from ppb to molecules/cc (air)
	# seedi - index of seed component(s)
	# diff_vol - diffusion volumes of components according to 
	#	Fuller et al. (1969)
	# DStar_org - gas-phase diffusion coefficients of components (cm2/s)
	# -----------------------------------------------------------------------

	# check on change of light setting --------------------------------------

	# begin by assuming no change to time interval required due to chamber 
	# condition/nucleation
	bc_red = 0
	
	if (len(light_time))>0:
	
		# whether lights on (1) or off (0) during this step
		lightm = light_stat[int(sum(light_time<=sumt)-1)]
		
		# check whether changes occur at start of this time step
		if (sumt == light_time[light_time_cnt] and light_time_cnt>-1):
			
			if (light_time_cnt<(len(light_stat)-1)):
				light_time_cnt += 1 # keep count of light setting index
			else:
				light_time_cnt = -1 # reached end
			# reset flag for time step reduction due to chamber condition
			bc_red = 0
				
		
		# check whether light on/off changes during proposed integration time step
		if (sumt+tnew > light_time[light_time_cnt] and light_time_cnt != -1):
			# if yes, then reset integration time step so that next step coincides 
			# with change
			tnew = light_time[light_time_cnt]-sumt
			bc_red = 1 # flag for time step reduction due to boundary conditions
			
		# if reached final status of lights, then keep this status
		if light_time_cnt == -1:
			lightm = light_stat[light_time_cnt]
 
	# if lights are on during this step and lighting is natural, then check whether
	# proposed time step needs reducing to limit change to light intensity, if this
	# time interval adaption is requested
	# if using natural light
	cwd = os.getcwd() # address of current working directory

	if (lightm == 1 and photo_par_file == str(cwd + '/PyCHAM/photofiles/MCMv3.2') and act_flux_path == 'no' and light_ad == 1):	
		# check time step required to limit change to rate of 
		# MCM photochemical equation number 6 (python index 5), 
		# which the unit test for
		# the zenith module shows to be most photosensitive (compared to
		# other photochemical equations)
		import zenith
		# photochemical rate now
		(secxn, cosxn) = zenith.zenith(sumt, lat, lon, dayOfYear)
		Jn =1.747E-01*cosxn**(0.155)*np.exp(-1.0*0.125*secxn)
		# photochemical rate after proposed time step
		(secxt, cosxt) = zenith.zenith(sumt+tnew, lat, lon, dayOfYear)
		Jt =1.747E-01*cosxn**(0.155)*np.exp(-1.0*0.125*secxn)
		# iteratively reduce proposed time interval until photochemical
		# rate changes by acceptable amount
		while (abs(Jt-Jn) > 5.e-3):
			tnew = tnew*0.9
			# photochemical rate after proposed time step
			(secxt, cosxt) = zenith.zenith(sumt+tnew, lat, lon, dayOfYear)
			Jt =1.747E-01*cosxt**(0.155)*np.exp(-1.0*0.125*secxt)
			bc_red = 1

	# check on updates to temperature (K) --------------------------------------	
	if len(temp)>1: # because a temperature must be given for experiment start
	
		# check whether changes occur at start of this time step
		if (sumt == tempt[temp_count]):

			# new temperature (K)
			temp_now = temp[temp_count]
			
			# update vapour pressure of water (log10(atm)), but don't change 
			# gas-phase concentration because we assume RH allowed to change with
			# varying temperature
			[_, Psat_water, _] = water_calc(temp_now, RH, 6.02214129e+23)
			
			# update vapour pressures of all components (molecules/cc and Pa), 
			# ignore density output
			[Psat, _, Psat_Pa] = volat_calc.volat_calc(0, Pybel_objects, temp_now, H2Oi,   
							num_comp, Psat_water, [], [], 0, corei, seed_name, 
							pconc, 0, 0.0, [], 1)

			# note, assume that air pressure inside chamber stays constant despite
			# varying temperature, therefore total molecular concentration must vary
			# (molecules/cc (air))
			ntot = Pnow*(NA/(si.R*temp_now))
			# remember number of molecules in one billionth of total number prior
			# to update
			Cfactor0 = Cfactor
			# update number of molecules in one billionth of this
			Cfactor = ntot*1.e-9 # ppb-to-molecules/cc
			
			# dynamic viscosity of air (kg/m.s), eq. 4.54 of Jacobson 2005
			dyn_visc = 1.8325e-5*((416.16/(temp_now+120.))*(temp_now/296.16)**1.5)
	
			ma = 28.966e-3 # molecular weight of air (kg/mol) (Eq. 16.17 Jacobson 2005)
			
			# air density (kg/m3 (air)), ideal gas law
			rho_a =  (Pnow*ma)/((si.R)*temp_now)
						
			# update mean free path and thermal speed
			# mean thermal speed of each molecule (m/s) (11.151 Jacobson 2005)
			# note that we need the weight of one molecule, which is why y_mw is divided by
			# Avogadro's constant, and we need it in kg, which is why we multiply by 1e-3
			therm_sp = ((8.*si.k*temp_now)/(np.pi*(y_mw/si.N_A)*1.e-3))**0.5
			
			# mean free path (m) for each component (15.24 of Jacobson 2005)
			# molecular weight of air (28.966 g/mol taken from table 16.1 Jacobson 2005)
			mfp = (2.*dyn_visc/(rho_a*therm_sp)).reshape(-1, 1)
			
			# diffusion coefficient (m2/s) of components in gas phase (air), eq 4.1.4 of
			# the Taylor (1993) textbook 
			# Multicomponent Mass Transfer, ISBN: 0-471-57417-1, note diffusion 
			# volume for air (19.7) taken from Table 4.1 of Taylor (1993) and mw of 
			# air converted to g/mol from kg/mol.  This is a replication of the original method 			# from Fuller et al. (1969): doi.org/10.1021/j100845a020
			DStar_org = 1.013e-2*temp_now**1.75*(((y_mw+ma*1.e3)/(y_mw*ma*1.e3))**0.5)/(Pnow*(diff_vol**(1./3.)+19.7**(1./3.))**2.)
			# convert to cm2/s
			DStar_org = DStar_org*1.e4
			
			
			# alter constant concentration (molecules/cc) of any components
			# with constant gas-phase concentration (ppb)
			if num_const_compi>0:
				
				y[const_compi[:]] = y[const_compi[:]]*(Cfactor/Cfactor0)

			if (temp_count<(len(tempt)-1)):
				temp_count += 1 # keep count of temperature setting index
			else:
				temp_count = -1 # reached end
			bc_red = 0 # reset flag for time step reduction due to boundary conditions
			
		# check whether temperature changes during proposed integration time step
		if (sumt+tnew > tempt[temp_count] and temp_count!=-1):
			# if yes, then reset integration time step so that next step coincides 
			# with change
			tnew = tempt[temp_count]-sumt
			bc_red = 1 # flag for time step reduction due to boundary conditions

	if (len(temp) == 1):
		temp_now = temp[0] # temperature constant if only one value given

	# check on instantaneous injection of components ---------------------------------------
	if len(injectt)>0 and gasinj_cnt>-1: # if any injections occur
	
		# check whether changes occur at start of this time step
		if (sumt == injectt[gasinj_cnt]):
			# account for change in gas-phase concentration,
			# convert from ppb/s to molecules/cc.s (air)
			y[inj_indx] += Ct[:, gasinj_cnt]*Cfactor-y[inj_indx]
			if (gasinj_cnt<(Ct.shape[1]-1)):
				gasinj_cnt += 1 # update count on injections
			else:
				gasinj_cnt = -1 # reached end
			bc_red = 0 # reset flag for time step reduction due to boundary conditions
				
		# check whether changes occur during proposed integration time step
		if (sumt+tnew > injectt[gasinj_cnt] and gasinj_cnt != -1):
			# if yes, then reset integration time step so that next step coincides 
			# with change
			tnew = injectt[gasinj_cnt]-sumt
			bc_red = 1 # flag for time step reduction due to boundary conditions
	
	# check on instantaneous injection of particles --------------------------------------	
	if ((sum(pconct[0, :]) > 0) and (seedt_cnt > -1) and (num_sb-wall_on > 0)): # if constant influx occurs
	
		# check whether changes occur at start of this time step
		if (sumt == pconct[0, seedt_cnt]):
			
			# account for change in seed particles
			[y[num_comp:num_comp*(num_sb-wall_on+1)], N_perbin, x, 
					Varr] = pp_dursim.pp_dursim(y[num_comp:num_comp*(num_sb-wall_on+1)], 
					N_perbin, 
					mean_rad[:, seedt_cnt], pmode, pconc[:, seedt_cnt], seedi, seedVr, lowsize, 
					uppsize, num_comp, (num_sb-wall_on), MV, rad0, radn, 
					std[:, seedt_cnt], y_dens, H2Oi, rbou)
			if (seedt_cnt<(pconct.shape[1]-1)):
				seedt_cnt += 1
			else:
				seedt_cnt = -1 # reached end
			bc_red = 0 # reset flag for time step reduction due to boundary conditions
			
		# check whether changes occur during proposed integration time step
		if (sumt+tnew > pconct[0, seedt_cnt] and seedt_cnt!=-1): 
			# if yes, then reset integration time step so that next step coincides 
			# with change
			tnew = pconct[0, seedt_cnt]-sumt
			bc_red = 1 # flag for time step reduction due to boundary conditions

	# check on continuous influxes of components ----------------------------------------------
	if (len(const_infl_t) > 0): # if influx occurs
	
		# in case influxes begin after simulation start create a zero array of correct shape
		if (sumt == 0. and const_infl_t[infx_cnt] != 0.):
			Cinfl_now = np.zeros((con_infl_C.shape[0], 1))
		
		# if the final input for influxes reached
		if (infx_cnt == -1):
			# influx of components now, convert from ppb/s to molecules/cc.s (air)
			Cinfl_now = (Cinfl[:, infx_cnt]*Cfactor).reshape(-1, 1)
			
		# check whether changes occur at start of this time step
		if (sumt == const_infl_t[infx_cnt] and (infx_cnt != -1)):
			
			# influx of components now, convert from ppb/s to molecules/cc.s (air)
			Cinfl_now = (Cinfl[:, infx_cnt]*Cfactor).reshape(-1, 1)
			
			# update index counter for constant influxes - used in integrator below
			if (infx_cnt<(Cinfl.shape[1]-1)):
				infx_cnt += 1
			else:
				infx_cnt = -1 # reached end
			bc_red = 0 # reset flag for time step reduction due to boundary conditions
			
		# check whether changes occur during proposed integration time step
		if (sumt+tnew > const_infl_t[infx_cnt] and (infx_cnt != -1)):
			# if yes, then reset integration time step so that next step coincides 
			# with change
			tnew = const_infl_t[infx_cnt]-sumt
			bc_red = 1 # flag for time step reduction due to boundary conditions
			
	else: # if no influxes, provide filler
		Cinfl_now = np.zeros((1, 1))
	
	# check on nucleation ---------------------------------------------------------
	# if automatic time step adaption to nucleation requested, check whether number of new particles
	# exceeds 10 % of total number formed during nucleation event.  Second part of condition is that
	# the specifiec nucleation event has not yet reached its defined finishing particle number
	# concentration (#/cc (air))
	if ((nuc_ad == 1) and (new_part_sum1 < nucv1*0.9) and ((num_sb-wall_on) > 0)):
		# the time step (s) needed to increase number concentration of nucleated particles by 10 %
		t_need = (0.1*nucv1+new_part_sum1)
		t_need = np.log(t_need/nucv1)
		t_need = np.log(t_need/nucv2)
		t_need = t_need*nucv3*-1.-sumt
	
		if tnew>t_need: # if suggested time step exceeds this, then reduce to required time step 
			tnew = t_need
			update_stp = t_need
			update_count = 0.
			bc_red = 1
			
	# nucleation check end -------------------------------------------------------------------------	


	return(temp_now, Pnow, lightm, light_time_cnt, tnew, bc_red, update_stp, update_count, 
		Cinfl_now, seedt_cnt, Cfactor, infx_cnt, gasinj_cnt, DStar_org)



'''The module that generates the Graphical User Interface for PyCHAM, and connects that GUI with the core PyCHAM model'''
# first module called when PyCHAM started from the terminal/command window, takes inputs
# and sends to model modules, also calls the saving module

from PyQt5.QtWidgets import QApplication, QWidget, QPushButton, QFileDialog, QLabel
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import pyqtSlot, Qt
import pickle # for storing inputs
import sys
import os
import def_mod_var
import numpy as np

class PyCHAM(QWidget):

	def __init__(self):
		super().__init__()
		self.title = 'PyCHAM'
		self.left = 10
		self.top = 10
		self.width = 450
		self.height = 300
		self.initUI()
		
		# default variables for all required input model variables 
		# stored to pickle file and output here
		[sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O] = def_mod_var.def_mod_var(0)
		return
    
	def initUI(self):
		self.setWindowTitle(self.title)
		self.setGeometry(self.left, self.top, self.width, self.height)
		
		label = QLabel(self)
		label.setText("Welcome to PyCHAM.  Please see the README file for guidance.")
		label.move(20, 35)
		label.show()
		
		button = QPushButton('Chemical Scheme File', self)
		button.setToolTip('Select the file containing the desired chemical scheme')
		button.move(100, 70)
		button.clicked.connect(self.on_click1)
		
		button = QPushButton('Chemical Scheme .xml File', self)
		button.setToolTip('Select the .xml file containing the desired conversion file')
		button.move(100, 100)
		button.clicked.connect(self.on_click2)
		
		button = QPushButton('Model Variables .txt File', self)
		button.setToolTip('Select the desired file containing the model variables')
		button.move(100, 130)
		button.clicked.connect(self.on_click3)
		
		button = QPushButton('Run Model', self)
		button.setToolTip('Start the simulation')
		button.move(100, 160)
		button.clicked.connect(self.on_click4)
		
		button = QPushButton('Plot Results', self)
		button.setToolTip('Plot output from simulation')
		button.move(100, 190)
		button.clicked.connect(self.on_click5)

		button = QPushButton('Quit', self)
		button.setToolTip('Finish with PyCHAM and close this window')
		button.move(100, 220)		
		button.clicked.connect(self.on_click6)
		
		self.show()
		return
		

	@pyqtSlot()
	def on_click1(self):
		# prepare by opening default inputs, ready for modification
		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'rb') as pk:
			[sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, 
			tot_time, comp0, y0, temp, tempt, RH, Press, wall_on,
			Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, 
			save_step, const_comp, Compt, injectt, Ct, seed_name,
			seed_mw, seed_diss, seed_dens, seedVr,
			light_stat, light_time, daytime, lat, lon, af_path, 
			dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, 
			dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, 
			accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, 
			nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, 
			inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O] = pickle.load(pk)
			pk.close()
			
		sch_name = self.openFileNameDialog() # get location of chemical scheme file
		# enforced file selection example
		#sch_name = str(os.getcwd() + '/PyCHAM/output/GMD_paper_plotting_scripts/fig01_scheme.txt')
		
		# pickle with new chemical scheme name	
		list_vars = [sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O]
		with open(input_by_sim, 'wb') as pk:
			pickle.dump(list_vars, pk) # pickle
			pk.close() # close

	@pyqtSlot()
	def on_click2(self):
		# prepare by opening default inputs, ready for modification
		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'rb') as pk:
			[sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, 
			tot_time, comp0, y0, temp, tempt, RH, Press, wall_on,
			Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, 
			save_step, const_comp, Compt, injectt, Ct, seed_name,
			seed_mw, seed_diss, seed_dens, seedVr,
			light_stat, light_time, daytime, lat, lon, af_path, 
			dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, 
			dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, 
			accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, 
			nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, 
			inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O] = pickle.load(pk)
			pk.close()
		xml_name = self.openFileNameDialog() # get location of xml file
			
		# pickle with new xml file name	
		list_vars = [sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O]
		with open(input_by_sim, 'wb') as pk:
			pickle.dump(list_vars, pk) # pickle
			pk.close() # close		

	@pyqtSlot()
	def on_click3(self):

		# prepare by opening default inputs, ready for modification
		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'rb') as pk:
			[sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, 
			tot_time, comp0, y0, temp, tempt, RH, Press, wall_on,
			Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, 
			save_step, const_comp, Compt, injectt, Ct, seed_name,
			seed_mw, seed_diss, seed_dens, seedVr,
			light_stat, light_time, daytime, lat, lon, af_path, 
			dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, 
			dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, 
			accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, 
			nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, 
			inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O] = pickle.load(pk)
			pk.close() # close pickle file


		inname = self.openFileNameDialog() # get location of model variables file
		# enforced selection example
		#inname = str(os.getcwd() + '/PyCHAM/output/GMD_paper_plotting_scripts/fig01_mod_var.txt')
		
		inputs = open(inname, mode= 'r' ) # open model variables file
		in_list = inputs.readlines() # read file and store everything into a list
		inputs.close() # close file
			
		for i in range(len(in_list)): # loop through supplied model variables to interpret

			key, value = in_list[i].split('=') # split values from keys
			# model variable name - a string with bounding white space removed
			key = key.strip()

			if key == 'res_file_name' and (value.strip()): # name of folder to save results in
				sav_nam = str(value.strip())

			if key == 'chem_scheme_markers' and (value.strip()): # formatting for chemical scheme
				chem_sch_mark = [str(i).strip() for i in (value.split(','))]

			if key == 'update_step' and (value.strip()): # time step (s) for updating ODE initial conditions
				update_stp = float(value.strip())

			if key == 'total_model_time' and (value.strip()):
				tot_time = float(value.strip())

			if key == 'Comp0' and (value.strip()): # names of components present at experiment start
				comp0 = [str(i).strip() for i in (value.split(','))]			

			if key == 'C0' and (value.strip()): # initial concentrations of components present at experiment start (ppb)
				y0 = [float(i) for i in (value.split(','))]

			if key == 'temperature' and (value.strip()): # chamber temperature (K)
				temp = [float(i) for i in ((value.strip()).split(','))]

			if key == 'tempt' and (value.strip()): # times (s) that temperature values correspond to
				tempt = [float(i) for i in ((value.strip()).split(','))]

			if key == 'rh' and (value.strip()): # relative humidity in chamber (0-1)
				RH = float(value.strip())

			if key == 'p_init' and (value.strip()): # pressure inside chamber
				Press = float(value.strip())

			if key == 'daytime_start' and (value.strip()): # time of day at experiment start (s)
				daytime = float(value.strip())

			if key == 'wall_on' and (value.strip()): # marker for whether or not to consider wall
				wall_on = int(value.strip())

			if key == 'eff_abs_wall_massC' and (value.strip()): # effective absorbing mass concentration of wall
				Cw = float(value.strip())

			if key == 'mass_trans_coeff' and (value.strip()): # mass transfer coefficient of vapours with wall
				kw = float(value.strip())

			if key == 'size_structure' and (value.strip()): # the size structure
				siz_stru = int(value.strip())

			if key == 'number_size_bins' and (value.strip()): # number of particle size bins
				num_sb = int(value.strip())

			if key == 'pconc' and (value.strip()): # seed particle number concentrations (#/cc)
				time_cnt = 1 # track number of times
				sb_cnt = 1 # track number of size bins
				mode_cnt = 1 # track number of modes
				for i in value:
					if i==';':
						time_cnt += 1 # increase time count
					if (time_cnt == 1 and i==','):
						sb_cnt += 1 # increase size bin count
						pmode = 1 # explicitly stated particle concentrations
					if (time_cnt == 1 and i == ':'):
						mode_cnt += 1 # mode count
						pmode = 0 # particle concentrations expressed as modes

				# if number concentration per size bin given explicitly
				if (sb_cnt > 1):
					pconc = np.zeros((sb_cnt, time_cnt))
					for i in range(time_cnt):
						pconc[:, i] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(','))]
				else: # mode quantities provided
					pconc = np.zeros((mode_cnt, time_cnt))
					for i in range(time_cnt):
						pconc[:, i] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(':'))]
				
			
			if key == 'pconct' and (value.strip()): # seed particle input times (s)
				time_cnt = 1 # track number of times
				for i in value:
					if (i == ';'):
						time_cnt += 1 # increase time count
					
				# times in columns
				pconct = np.zeros((1, time_cnt))
				pconct[0, :] = [float(i) for i in ((value.strip()).split(';'))]

			if key == 'lower_part_size' and (value.strip()): # lowest size bin bound
				lowsize = float(value.strip())
			
			if key == 'upper_part_size' and (value.strip()): # radius of uppermost size bin boundary
				uppsize = float(value.strip())

			if key == 'space_mode' and (value.strip()): # method of spacing size bin sizes
				space_mode = str(value.strip())

			if key == 'std' and (value.strip()): # seed particle standard deviation of number size distribution
				time_cnt = 1 # track of number of times
				mode_cnt =1 # track number of modes 
				for i in value:
					if i==';':
						time_cnt += 1 # increase time count
					if i==':':
						mode_cnt += 1 # increase mode count
				std = np.zeros((mode_cnt, time_cnt))
				for i in range(time_cnt):
					std[:, i] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(':'))]
			
			if key == 'mean_rad' and (value.strip()): # seed particle mean radius (um)
				time_cnt = 1 # track of number of times
				mode_cnt = 1 # track number of modes
				for i in value:
					if i==';':
						time_cnt += 1 # increase time count
					if i==':':
						mode_cnt += 1 # increase mode count
				
				mean_rad = np.zeros((mode_cnt, time_cnt))
				for i in range(time_cnt):
					mean_rad[:, i] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(':'))]
				
			if key == 'recording_time_step' and (value.strip()): # frequency (s) of storing results
				save_step = float(value.strip())

			if key == 'const_comp' and (value.strip()): # names of components with later continuous injections
				const_comp = [str(i).strip() for i in (value.split(','))]
			
			# names of components with instantaneous gas-phase injections
			if key == 'Compt' and (value.strip()):
				Compt = [str(i).strip() for i in (value.split(','))]

			# times of later gas-phase instantaneous injections (s)
			if key == 'injectt' and (value.strip()):
				injectt = [float(i.strip()) for i in (value.split(','))]
				injectt = np.array((injectt))

			# concentration of components with instantanetous injection (ppb)
			if key == 'Ct' and (value.strip()):
				comp_count = 1 # count number of components
				time_count = 1 # track number of times
				for i in value:
					if i==';':
						comp_count += 1 # record number of components
						time_count = 1 # reset time count
					if i==',':
						time_count += 1
				Ct = np.zeros((comp_count, time_count))
				for i in range(comp_count):
					Ct[i, :] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(','))]


			if key == 'seed_name' and value.strip(): # name(s) of component(s) comprising seed particles
				seed_name = [str(i).strip() for i in (value.split(','))]

			if key == 'seed_mw' and value.strip():
				seed_mw = float(value.strip())
		
			if key == 'seed_diss' and value.strip(): # dissociation constant of seed material
				seed_diss = [float(i) for i in (value.split(','))]

			if key == 'seed_dens' and value.strip():
				seed_dens = float(value.strip())

			if key == 'seedVr' and value.strip(): # volume ratio of components in seed particles
				seedVr = [float(i) for i in (value.split(','))]

			if key == 'light_status' and value.strip(): # status of lights (on or off)
				light_stat = [int(i) for i in (value.split(','))]
				light_stat = np.array((light_stat))
				
			if key == 'light_time' and value.strip(): # times (s) corresponding to light status
				light_time = [float(i) for i in (value.split(','))]
				light_time = np.array((light_time))
				
			if key == 'lat': # latitude (degrees)
				if (value.strip()): lat = float(value.strip())

			if key == 'lon': # longitude (degrees)
				if (value.strip()): lon = float(value.strip())

			if key == 'act_flux_file' and (value.strip()): # for indoor actinic flux
				af_path = str(os.getcwd() + '/PyCHAM/photofiles/' + value.strip())

			if key == 'DayOfYear' and (value.strip()):
				dayOfYear = int(value.strip())		
			
			# name of file with wavelength-dependent absorption cross-section and quantum yield 
			# calculations
			if key == 'photo_par_file' and (value.strip()):
				photo_path = str(os.getcwd() + '/PyCHAM/photofiles/' + value.strip())
			
			if key == 'trans_fac' and (value.strip()): # transmission factor for natural light
				tf = float(value.strip())

			if key == 'light_adapt' and (value.strip()): # whether to adapt time step to naturally varying light intensity
				light_ad = int(value.strip())

			if key == 'const_infl' and (value.strip()): # names of components with continuous influx
				con_infl_nam = [str(i).strip() for i in (value.split(','))]

			if key == 'const_infl_t' and (value.strip()): # times of continuous influxes (s)
				con_infl_t = [float(i.strip()) for i in (value.split(','))]
				con_infl_t = np.array((con_infl_t))

			if key == 'Cinfl' and (value.strip()): # influx rate of components with continuous influx (ppb/s)
				comp_count = 1 # count number of components
				time_count = 1 # track number of times
				for i in value:
					if i==';':
						comp_count += 1 # record number of components
						time_count = 1 # reset time count
					if i==',':
						time_count += 1
				con_infl_C = np.zeros((comp_count, time_count))
				for i in range(comp_count):
					con_infl_C[i, :] = [float(ii.strip()) for ii in ((value.split(';')[i]).split(','))]

			if key == 'tracked_comp' and (value.strip()): # names of components whose tendency to change will be tracked
				dydt_trak = [str(i).strip() for i in (value.split(','))]

			if key == 'dens_Comp' and (value.strip()):
				dens_comp = [str(i).strip() for i in (value.split(','))]

			if key == 'dens' and (value.strip()):
				dens = [float(i) for i in (value.split(','))]

			if key == 'vol_Comp' and (value.strip()):
				vol_comp = [str(i).strip() for i in (value.split(','))]

			if key == 'volP' and (value.strip()):
				volP = [float(i) for i in (value.split(','))]

			if key == 'act_comp' and (value.strip()): # names of componentes with specified activity coefficients
				act_comp = [i for i in (((value.strip()).split(',')))]

			if key == 'act_user' and (value.strip()): # activity coefficients (dimensionless set by user)
				act_user = [i for i in (((value.strip()).split(',')))]

			if key == 'accom_coeff_comp' and (value.strip()): # names of componenets with specified accommodation coefficients
				accom_comp = [i for i in (((value.strip()).split(',')))]

			if key == 'accom_coeff_user' and (value.strip()): # value(s) of accommodation coefficient set by user
				accom_val = [i for i in (((value.strip()).split(',')))]

			if key == 'partit_cutoff' and (value.strip()): # value(s) of the gas-particle partitioning cutoff
				partit_cutoff = [float(i) for i in (((value.strip()).split(',')))]

			if key == 'umansysprop_update' and (value.strip()): # marker for whether to clone latest version of UManSysProp from web
				uman_up = int(value.strip())

			if key == 'int_tol' and (value.strip()): # tolerances for integration
				int_tol = [float(i) for i in (value.split(','))]

			if key == 'new_partr' and (value.strip()): # radius of newly nucleated particles (cm)
				new_partr = float(value.strip())

			if key == 'nucv1' and (value.strip()): # first parameter in the nucleation equation
				nucv1 = float(value.strip())

			if key == 'nucv2' and (value.strip()): # second nucleation parameter (onset)
				nucv2 = float(value.strip())

			if key == 'nucv3' and (value.strip()): # third nucleation parameter (duration)
				nucv3 = float(value.strip())

			if key == 'nuc_comp' and (value.strip()): # chemical scheme name of nucleating component
				nuc_comp = [str(i).strip() for i in (value.split(','))]

			if key == 'nuc_adapt' and (value.strip()): # marker for whether to adapt time interval to nucleation
				nuc_ad = int(value.strip())

			if key == 'coag_on' and (value.strip()): # marker for whether to model coagulation
				coag_on = int(value.strip())

			if key == 'inflectDp' and (value.strip()): # diameter at which wall deposition of particles inflection occurs
				inflectDp = float(value.strip())

			if key == 'Grad_pre_inflect' and (value.strip()): # gradient of wall deposition of particles before inflection
				pwl_xpre = float(value.strip())

			if key == 'Grad_post_inflect' and (value.strip()): # gradient of particle deposition to wall after inflection
				pwl_xpro = float(value.strip())

			if key == 'Rate_at_inflect' and (value.strip()): # particle deposition to wall rate at inflection
				inflectk = float(value.strip())

			if key == 'ChamSA' and (value.strip()): # chamber surface area (m2) used for particle loss to walls
				ChamSA = float(value.strip())

			if key == 'McMurry_flag' and (value.strip()): # marker for whether to use the McMurry model for particle deposition to wall
				Rader = int(value.strip())

			if key == 'part_charge_num' and (value.strip()): # average number of charges per particle
				p_char = float(value.strip())
			
			if key == 'elec_field' and (value.strip()): # electric field in chamber
				e_field = float(value.strip())
			
			if key == 'dil_fac' and (value.strip()): # dilution factor rate
				dil_fac = float(value)
				
			if key == 'ser_H2O' and (value.strip()): # whether to serialise water gas-particle partitioning
				ser_H2O = int(value)
		
		# prepare for pickling
		list_vars = [sav_nam, sch_name, chem_sch_mark, xml_name, update_stp, tot_time, comp0, y0, temp, tempt, RH, Press, wall_on, Cw, kw, siz_stru, num_sb, pmode, pconc, pconct, lowsize, uppsize, space_mode, std, mean_rad, save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, seed_diss, seed_dens, seedVr, light_stat, light_time, daytime, lat, lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, con_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, volP, act_comp, act_user, accom_comp, accom_val, uman_up, int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, inflectDp, pwl_xpre, pwl_xpro, inflectk, chamSA, Rader, p_char, e_field, dil_fac, partit_cutoff, ser_H2O]

		input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
		with open(input_by_sim, 'wb') as pk: # the file to be used for pickling
			pickle.dump(list_vars, pk) # pickle
			pk.close() # close


		
	@pyqtSlot()	
	def on_click4(self): # button to run simulation
		import middle
		# call on modules to solve problem
		middle.middle()
	
	@pyqtSlot() # button to plot results graphically
	def on_click5(self):	
		import plotter
		print('Plotting and saving standard results graph')
		plotter.plotter(0) # plot results

	@pyqtSlot() # button to quit software
	def on_click6(self):
		QWidget.close(self)
		
	def openFileNameDialog(self): # allows opening of system's directory navigator
		options = QFileDialog.Options()
		fname, _ = QFileDialog.getOpenFileName(self,"QFileDialog.getOpenFileName()", "","All Files (*);;Python Files (*.py)", options=options)
		return(fname)



'''module to calculate Reynold number'''

import numpy as np
import scipy.constants as si

def Reyn_num(sbr, eta_a, rho_a, kin_visc, sbrho, Kni):

	# --------------------------------------------------------
	# inputs:
	# sbr - particle radius (m)
	# eta_a - dynamic viscosity of air (4.54) (g/m.s)	
	# rho_a - density of air (g/m3)
	# kin_visc - kinematic viscosity of air (m2/s)
	# sbrho - average density of particles (g/m3)
	# Kni - Knudsen number (dimensionless)

	# --------------------------------------------------------
	# outputs:
	# Re - Reynold number (dimensionless)
	# Vf - terminal fall speed (m/s)
	# --------------------------------------------------------


	# Cunningham slip-flow correction (15.30) with constant taken
	# from text below
	G = 1.0+Kni*(1.249+0.42*(np.exp(-0.87/Kni)))
	
	# terminal fall speed (m/s) (20.4)
	Vf = ((2.0*sbr**2.0*(sbrho-rho_a)*si.g)/(9.0*eta_a))*G
	# particle Reynolds number (15.26) (dimensionless)
	Re = (2.0*sbr*Vf)/kin_visc
	
	return Re, Vf




'''estimate of the diffusion volume of components'''
# following the method of Fuller et al. (1966) (doi.org/10.1021/ie50677a007) and Fuller et al. (1969) (doi.org/10.1021/j100845a020), the calculated, with the volume increments given in Table 4.1 
# of the Taylor (1993) textbook Multicomponent Mass Transfer, ISBN: 0-471-57417-1
# diffusion volumes will be used to estimate gas-phase diffusion coefficients
# documents for pybel: http://openbabel.org/docs/dev/UseTheLibrary/Python_PybelAPI.html#pybel.Molecule.sssr

import numpy as np
import pybel
import math

# define function
def diff_vol_est(Pybel_object):

	# inputs: -------------------------------------------------------------------------
	# Pybel_object - Pybel objects of components
	# -----------------------------------------------------------------------------------
	
	# simple molecules given in Table 4.1 of Taylor (1993)
	# SMILE strings for simple molecules
	simp_molec = ['[He]', '[Ne]', '[Ar]', '[Kr]', '[Xe]', '[H][H]', r'N#N', 'O=O', 'CO', 'O=C=O', 'NO=N', '[H]N([H])[H]', 'O', 'S(F)(F)(F)(F)(F)F', 'ClCl', 'BrBr', 'O=S=O']
	
	# diffusion volumes for simple molecules given in Table 4.1 of  the Taylor (1993) textbook 
	# Multicomponent Mass Transfer, ISBN: 0-471-57417-1
	simp_molec_vol = [2.67, 5.98, 16.2, 24.5, 32.7, 6.12, 18.5, 16.3, 18., 26.7, 35.9, 20.7, 13.1, 71.3, 38.4, 69., 41.8]
	
	Pybel_object_ref = [] # empty list
	# convert to SMARTS (pybel_objects)
	# generate pybel objects from SMILES
	for i in simp_molec:	
		Pybel_object_ref.append(pybel.readstring('smi', i))
	
	diff_vol = np.zeros((len(Pybel_object))) # empty results array
	
	for compi in range(len(Pybel_object)): # component loop
	
		simp_cnt = 0 # count
	
		# check if component is a simple molecule
		for i_smrt in Pybel_object_ref: # loop through simple molecules
			if (Pybel_object[compi].formula == i_smrt.formula):
				diff_vol[compi] = simp_molec_vol[simp_cnt]
				break
			simp_cnt += 1
			
		# if a simple molecule move onto next component
		if (simp_cnt < len(Pybel_object_ref)):
			continue
			
		# SMARTS for the atoms given in Table 4.1 of  the Taylor (1993) textbook 
		# Multicomponent Mass Transfer, ISBN: 0-471-57417-1
		smarts = []
		smarts.append(pybel.Smarts('[#6]'))
		smarts.append(pybel.Smarts('[H]'))
		smarts.append(pybel.Smarts('[#8]'))
		smarts.append(pybel.Smarts('[#7]'))
		smarts.append(pybel.Smarts('[#16]'))
		smarts.append(pybel.Smarts('[#9]'))
		smarts.append(pybel.Smarts('[#17]'))
		smarts.append(pybel.Smarts('[#35]'))
		smarts.append(pybel.Smarts('[#53]'))
	
		# count atoms present, final element will be used for number of rings
		atm_cnt = np.zeros((len(smarts)+1))
	
		# atomic and molecular diffusion volume increments (Table 4.1 Taylor 1993)
		vol_inc = np.array((15.9, 2.31, 6.11, 4.54, 22.9, 14.7, 21., 21.9, 29.8, -18.3))
	
		i_cnt = 0
		for i_smrt in smarts: # loop through reference groups
			atm_cnt[i_cnt] = len(i_smrt.findall(Pybel_object[compi]))
			i_cnt += 1
			
		# smarts for rings
		smarts = []
		smarts.append(pybel.Smarts('[r3]'))
		smarts.append(pybel.Smarts('[r4]'))
		smarts.append(pybel.Smarts('[r5]'))
		smarts.append(pybel.Smarts('[r6]'))
		smarts.append(pybel.Smarts('[r7]'))
		smarts.append(pybel.Smarts('[r8]'))
		smarts.append(pybel.Smarts('[r9]'))
		smarts.append(pybel.Smarts('[r10]'))
		
		# check for rings
		num_atm = 3 # count on number of atoms in rings
		for i_smrt in smarts: # loop through reference groups
			atm_cnt[i_cnt]  += math.ceil(len(i_smrt.findall(Pybel_object[compi]))/num_atm)
			num_atm += 1 # count on number of atoms in rings

		# diffusion volume
		diff_vol[compi] = np.sum((atm_cnt*vol_inc))

	return(diff_vol)


'''module to allow complete evaporation of the smallest size bin when conditions set in Vchange_check.py met'''
# rather than reduce the time step to unpractical length, allow the smallest size bin
# to completely evaporate using this module

def compl_evap(ytest, n0, Vnew, Vol0, nc, sbi):

	# inputs: ----------------------------------------
	# ytest - concentration (molecules/cc (air)) of components
	# n0 - number concentration of particles per size bin (# particles/cc (air))
	# Vnew - new estimates of single particle volume (um3)
	# Vol0 - default volume per size bin (um3), volume at centre of size bin bounds
	# nc - number of components
	# sbi - index of size bins in question
	# ------------------------------------------------

	# allow complete evaporation of the smallest size bin
	# components move to gas-phase (molecules/cc (air))
	ytest[0:nc] += ytest[(sbi+1)*nc:(sbi+2)*nc]
	# effectively zero concentration in particle-phase (molecules/cc (air))
	ytest[(sbi+1)*nc:(sbi+2)*nc] = 1.0e-40
	# remove particle concentration from size bin (# particles/cc (air))
	n0[sbi] = 1.0e-40
	# with no particles in smalles size bin now, default to volume at mid-point between 
	# size bin bounds (um3)
	Vnew[0] = Vol0[0]
	
	return(ytest, n0, Vnew)



'''generates an array of component indices for the components contained in the peroxy radical list'''
# makes a two column array, with the first giving the index of components
# included in the peroxy radical list based on component name and the
# second column containing the index based on component SMILE

import numpy as np

def write_RO2_indices(smiles_array, RO2_names):
    
    # store the names of RO2 species which are present in the equation file
    # get a list of INDICES of RO2 that present in the equation file 
    # (or total species dict)
    # empty list for RO2_indices
    RO2_indices0 = []
    RO2_indices = []
    
    for name in RO2_names:
        
        if name in smiles_array:
            # get the RO2 index
            index0 = RO2_names.index(name)
            RO2_indices0.append(index0)
            # get the smiles_array index for this RO2 species
            index1 = smiles_array.index(name)
            RO2_indices.append(index1)
    
    # Ensure elements in RO2_indices are int (iterable)
    RO2_indices0 = (np.asarray(RO2_indices0, dtype=int)).reshape(-1, 1)
    RO2_indices = (np.asarray(RO2_indices, dtype=int)).reshape(-1, 1)
    RO2_indices = np.hstack((RO2_indices0, RO2_indices))
    
    return (RO2_indices)
